
/**************************************************************************
 *                                                                        *
 * Copyright (c) 1996 Michael Richmond and Richard Treffers               *
 *                                                                        *
 *                    This software may be copied and distributed for     *
 *                    educational, research and not for profit services   *
 *                    provided that this copyright and statement are      *
 *                    included in all such copies.                        *
 *                                                                        *
 **************************************************************************/


/*	
	Clips data files

		CLIP file [MAX=] [VMAX=] [MIN=] [VMIN=] [box=d]

	sets values > max to VMAX and/or values < VMIN to VMIN
	operates over box if specified

	3/20/92 - nr=ncol bug fixed -RRT
	9/22/92 - usage 
			- exit(0) added -rrt
	7/28/93 - outfile added as well as verbose -rrt
	7/25/2011 - added "pmask=" and "pval=" options,
	               so that users can set pixels based
	               on their value in a pixel-mask.  MWR

	6/14/2013 - modified so if user specifies "outfile=", we copy
	                entire contents to new output file first,
	                then perform the requested action.  MWR

*/

#include <stdio.h>
#include <stdlib.h>
#include "pcvista.h"
#include "fits.h"
#include <unistd.h>
#include <string.h>

#ifdef PROTO
int main(int, char **);
#endif

#define USAGE \
  "usage: clip file [max=] [min=] [vmax=] [vmin=] [box=] [outfile=] \n       [pmask=] [pval=] [verbose] [help]"

int verbose;

int
main(argc, argv)
int argc;
char *argv[];
{
	FITS_HANDLE fh,fout, pmask_fh;
	int i, boxno;
	static int sr, sc, nr, nc;
	int nrow, ncol, row;
	int pmask_nrow, pmask_ncol;
	static int pmask_flag = 0;
	static int min = MIN_DATA_VAL, max = MAX_DATA_VAL, vmin = 0, vmax = 0, val;
	static int pval = 0;
	static int16 data[NMAX], clip_hi, clip_lo, got_vmin, got_vmax, pmask_data[NMAX];
	char *gotit;
	char outfile[NBUF], pmask_file[NBUF];
	char err_msg[NBUF];
	int new_file=0;

	if (argc == 1) {
		error(-1, USAGE);
	}
	if (strcmp(argv[1], "help") == 0) {
		printf("%s\n", USAGE);
		exit(0);
	}

	fh = fits_open(argv[1], "r", &nrow, &ncol);
	strcpy(outfile,argv[1]);
	nr = nrow;	
	nc = ncol;

	for (i = 2; i < argc; i++) {
		if (keyword("help", argv[i])) {
			printf("%s\n", USAGE);
			exit(0);
		}
		if ((gotit = find("box", argv[i])) != NULL) {
			boxno = (int)(evaluate(gotit) + 0.5);
			if (getbox(boxno, &sr, &sc, &nr, &nc))
				error(-1, "can't find box");		
			check_box(boxno, nrow, ncol);
			continue;
		}
		if ((gotit = find("min", argv[i])) != NULL) {
			min = (int) evaluate(gotit);
			clip_lo = 1;
			continue;
		}
		if ((gotit = find("vmin", argv[i])) != NULL) {
			vmin = (int) evaluate(gotit);
			got_vmin = 1;
			continue;
		}
		if ((gotit = find("max", argv[i])) != NULL) {
			max = (int) evaluate(gotit);
			clip_hi = 1;
			continue;
		}
		if ((gotit = find("outfile", argv[i])) != NULL) {
				strcpy(outfile,gotit);
				new_file=1;
				continue;
		}
		if ((gotit = find("pmask", argv[i])) != NULL) {
				strcpy(pmask_file,gotit);
				pmask_flag = 1;
				continue;
		}
		if ((gotit = find("pval", argv[i])) != NULL) {
				pval = (int) evaluate(gotit);
				continue;
		}

		if ((gotit = find("vmax", argv[i])) != NULL) {
			vmax = (int) evaluate(gotit);
			got_vmax = 1;
			continue;
		}
		if(keyword("verbose",argv[i])){
			verbose=1;
			continue;
		}
	}
	fout = fits_open(outfile, new_file?"w":"x", &nrow, &ncol);
		 
	/*
	 * If the user specifies "outfile=", then we do two steps:
	 *   1.  copy original file into new output file, as-is
	 *   2.  then, operate on the new output file as directed by user.
	 * If we don't perform a complete copy first, and user requests
	 * action only within a box, then the regions outside the box
	 * won't have valid data copied into them.
	 */
	if (new_file) {
		fits_copyheader(fh, fout, FITS_CHECK);
		for (row = 0; row < nrow; row++) {
			fits_get_data(fh, row, 0, data, ncol);
			fits_put_data(fout, row, 0, data, ncol);
		}
	}
			

	if ((pmask_flag == 0) && (pval != 0)) {
		error(-1, "pval= argument requires a pmask= argument");
	}

	if (pmask_flag == 1) {
		pmask_fh = fits_open(pmask_file, "r", &pmask_nrow, &pmask_ncol);
		if (nrow != pmask_nrow) {
			sprintf(err_msg, "number of rows in mask %s != number in %s \n",
					pmask_file, argv[1]);
			error(-1, err_msg);
		}
		if (ncol != pmask_ncol) {
			sprintf(err_msg, "number of cols in mask %s != number in %s \n",
					pmask_file, argv[1]);
			error(-1, err_msg);
		}
	}


	if (!got_vmax)
		vmax = max;
	if (!got_vmin)
		vmin = min;

	/* 
	 * note that if the user did give "pmask=", but did not supply
	 *   a value for "pval=", then we will set pval=0.  Thus,
	 *   all pixels with a mask value != 0 will have the values
	 *   in the output image set to zero.
	 */
	if ((pmask_flag == 1) && verbose) {
		printf("Setting masked pixels to     %10d \n", pval);
	}
	if (clip_hi && verbose)
		printf("Clipping values greater than %10d to %10d\n", max, vmax);
	if (clip_lo && verbose)
		printf("Clipping values less than    %10d to %10d\n", min, vmin);

	if (!clip_hi && !clip_lo && !pmask_flag)
		error(-1, "clip: no action taken");
	for (row = sr; row < (sr + nr); row++) {
		fits_get_data(fh, row, sc, data, nc);
		if (pmask_flag == 1) {
			fits_get_data(pmask_fh, row, sc, pmask_data, nc);
		}
		for (i = 0; i < nc; i++) {
			val = data[i];
			if ((pmask_flag == 1) && (pmask_data[i] != 0)) {
				data[i] = pval;
				continue;
			}
			if (clip_hi && (val > max)) {
				data[i] = vmax;
				continue;
			}
			if (clip_lo && (val < min)) {
				val = vmin;
				data[i] = val;		
			}
		}
		fits_put_data(fout, row, sc, data, nc);
	}
	fits_close(fout);
	if (pmask_flag == 1) {
		fits_close(pmask_fh);
	}
	exit(0);
}
