/**************************************************************************
 *                                                                        *
 * Copyright (c) 1996 Michael Richmond and Richard Treffers               *
 *                                                                        *
 *                    This software may be copied and distributed for     *
 *                    educational, research and not for profit services   *
 *                    provided that this copyright and statement are      *
 *                    included in all such copies.                        *
 *                                                                        *
 **************************************************************************/


/**************************************************************************
 * Compares two simple, 2-D 16-bin integer FITS images
 *     and decides if they are "identical", according to the
 *     following rules:
 *
 *     - must have the same number of rows and columns of data
 *     - all pixel values must be identical
 *     - all FITS header lines must be identical,
 *          EXCEPT for 'COMMENT' lines
 *
 * USAGE:  compfits  file1 file2  
 *
 * RETURNS:      0          if the two are "identical"
 *               1          if any tests fail
 *
 */

#include "pcvista.h"
#include "fits.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

#undef DEBUG


#define CARD_LEN     80      /* length of a FITS header line */

static int compare_nrow_ncol(int nrow1, int ncol1, int nrow2, int ncol2);

static int compare_pixel_values(FITS_HANDLE fh1, FITS_HANDLE fh2,
                                int nrow1, int ncol1, int nrow2, int ncol2);
static int compare_headers(FITS_HANDLE fh1, FITS_HANDLE fh2);
static int read_fits_key_val(char *line, char *keyword, char *value);


#define USAGE \
  "usage: compfits file1 file2  [help]"


int
main(argc,argv)
int argc;
char *argv[];
{
	FITS_HANDLE fh1, fh2;
	static int i, k, constant, boxno;
	static int div_const;
	static int nr1, nc1, sr1, sc1, nrow1, ncol1;
	static int nr2, nc2, sr2, sc2, nrow2, ncol2;
	static int16 data1[NMAX], data2[NMAX];
	char *gotit;
	char err_msg[NBUF];

	if ((argc == 1) || (argc < 3)) {
		fprintf(stderr, "%s\n", USAGE);
		exit(1);
	}
	if (strcmp(argv[1], "help") == 0) {
		printf("%s\n", USAGE);
		exit(0);
	}

   /* first arg must be FITS image number 1 */
	fh1 = fits_open(argv[1], "r", &nrow1, &ncol1);
	nr1 = nrow1;
	nc1 = ncol1;	
	sr1 = 0;
	sc1 = 0;

   /* second arg must be FITS image number 2 */
	fh2 = fits_open(argv[2], "r", &nrow2, &ncol2);
	nr2 = nrow2;
	nc2 = ncol2;	
	sr2 = 0;
	sc2 = 0;

   /* now parse any args after the two FITS images */
	for(i = 3; i < argc; i++) {
		if (keyword("help", argv[i])) {
			printf("%s\n", USAGE);
			exit(0);
		}
#if 0
		if ((gotit = find("box", argv[i])) != NULL) {
			boxno = (int)evaluate(gotit);
			if (getbox(boxno, &sr, &sc, &nr, &nc))
				error(-1, "can't find box");
			check_box(boxno, nrow1, ncol1);			
			check_box(boxno, nrow2, ncol2);			
			continue;
		}
#endif
	 }

   /*
	 * We now go through a series of tests.  As soon as any test fails,
	 *    we immediately stop, print error message, and terminate.
	 */
#ifdef DEBUG
	printf("compfits: comparing %s and %s \n", argv[1], argv[2]);
#endif

	if (compare_nrow_ncol(nrow1, ncol1, nrow2, ncol2) != 0) {
		return(1);
	}
	if (compare_pixel_values(fh1, fh2, nrow1, ncol1, nrow2, ncol2) != 0) {
		return(1);
	}
	if (compare_headers(fh1, fh2) != 0) {
		return(1);
	}


	fits_close(fh1);
	fits_close(fh2);


	exit(0);
}




/**********************************************************************
 * PROCEDURE: compare_nrow_ncol 
 *
 * DESCRIPTION: Compare the number of pixels in the two images.
 *              
 * RETURNS:   0        if the two have the same number of rows and cols
 *            1        otherwise
 *
 */

static int
compare_nrow_ncol
	(
	int nrow1,          /* I: number of rows in image 1 */
	int ncol1,          /* I: number of cols in image 1 */
	int nrow2,          /* I: number of rows in image 2 */
	int ncol2           /* I: number of cols in image 2 */
	)
{

#ifdef DEBUG
		printf("compare_nrow_ncol: comparing nrow1 %d nrow2 %d ", nrow1, nrow2);
#endif
	if (nrow1 != nrow2) {
#ifdef DEBUG
		printf("  fails \n");
#endif
		error(-1, "number of rows differs \n");
		return(1);
	}
	else {
#ifdef DEBUG
		printf("  okay \n");
#endif
	}

#ifdef DEBUG
		printf("compare_nrow_ncol: comparing ncol1 %d ncol2 %d ", ncol1, ncol2);
#endif
	if (ncol1 != ncol2) {
#ifdef DEBUG
		printf("  fails \n");
#endif
		error(-1, "number of cols differs \n");
		return(1);
	}
	else {
#ifdef DEBUG
		printf("  okay \n");
#endif
	}


	return(0);
}


/**********************************************************************
 * PROCEDURE: compare_pixel_values 
 *
 * DESCRIPTION: Compare the values of each pixel in two images.
 *              
 * RETURNS:   0        if the two have identical data in every pixel
 *            1        otherwise
 *
 */

static int
compare_pixel_values
	(
	FITS_HANDLE fh1,    /* I: handle to image 1 */
	FITS_HANDLE fh2,    /* I: handle to image 2 */
	int nrow1,          /* I: number of rows in image 1 */
	int ncol1,          /* I: number of cols in image 1 */
	int nrow2,          /* I: number of rows in image 2 */
	int ncol2           /* I: number of cols in image 2 */
	)
{
	int row, col;
	int sc, nc;
	int16 data1[NMAX], data2[NMAX];

	/* 
	 * might add additional arguments to check data within a box
	 *    in the future 
	 */
	sc = 0;
	nc = ncol1;

	/* sanity checks -- should not have to do this */
	if (nrow1 != nrow2) {
		printf("compare_pixel_values: nrow mismatch \n");
		return(1);
	}
	if (ncol1 != ncol2) {
		printf("compare_pixel_values: ncol mismatch \n");
		return(1);
	}

#ifdef DEBUG
	printf("compare_pixel_values: comparing pixel values \n");
#endif

	for (row = 0; row < nrow1; row++) {

		fits_get_data(fh1, row, sc, data1, nc);
		fits_get_data(fh2, row, sc, data2, nc);
		for (col = 0; col < ncol1; col++) {
			if (data1[col] != data2[col]) {
#ifdef DEBUG
				printf("  pixel (%5d,%5d) differs: %6d vs. %6d \n",
						row, col, data1[col], data2[col]);
#endif
				return(1);
			}
		}
	}

	/* if we get here, all pixel values were identical */
	return(0);
}



/**********************************************************************
 * PROCEDURE: compare_headers
 *
 * DESCRIPTION: Compare the FITS headers of two images.
 *              
 *              In order to count as "identical", the two 
 *              must have
 *
 *                a) the same number of header lines
 *                b) identical keyword and value items in all lines
 *                        which are not "COMMENT" lines
 *
 *              
 * RETURNS:   0        if the two have "identical" headers
 *            1        otherwise
 *
 */

static int
compare_headers
	(
	FITS_HANDLE fh1,    /* I: handle to image 1 */
	FITS_HANDLE fh2     /* I: handle to image 2 */
	)
{
	int card;
	int pair1, pair2;
	int16 data1[NMAX], data2[NMAX];
	char *gotit1, *gotit2;
	char key1[CARD_LEN + 1], val1[CARD_LEN + 1];
	char key2[CARD_LEN + 1], val2[CARD_LEN + 1];


#ifdef DEBUG
	printf("compare_headers: comparing header lines \n");
#endif

	for (card = 1; ; card++) {

		/* grab corresponding lines from each file */
		gotit1 = fits_get_headerline(fh1, card);
		gotit2 = fits_get_headerline(fh2, card);

		/* 
		 * if both lines are NULL, then both headers ended 
		 *    at the same location.  That's okay --
		 *    and so we can return with code 0
		 */
		if ((gotit1 == NULL) && (gotit2 == NULL)) {
			return(0);
		}

		/* if only one line is NULL, then they differ */
		if ((gotit1 == NULL) || (gotit2 == NULL)) {
			if (gotit1 == NULL) {
#ifdef DEBUG
				printf("  header1 ends before header2 \n");
#endif
				return(1);
			}
			if (gotit2 == NULL) {
#ifdef DEBUG
				printf("  header2 ends before header1 \n");
#endif
				return(1);
			}
		}

		/*
		 * if we reach this point, we've read something from each
		 *    of the two headers.  Let's see if the contents are
		 *    the same ...
		 */
		pair1 = read_fits_key_val(gotit1, key1, val1);
		pair2 = read_fits_key_val(gotit2, key2, val2);
		if ((pair1 == 0) && (pair2 == 0)) {
			/* 
			 * we read a keyword and value from each line.
			 *    Both keywords and values must match to count as "identical"
			 */
			if (strcmp(key1, key2) != 0) {
#ifdef DEBUG
				printf("    key mismatch: %s vs. %s \n", key1, key2);
#endif
				return(1);
			}
			if (strcmp(val1, val2) != 0) {
#ifdef DEBUG
				printf("    val mismatch for key %s: %s vs. %s \n", 
							key1, val1, val2);
#endif
				return(1);
			}

			/* if we get here, both keys and values were the same.  Fine */
#ifdef DEBUG2
			printf("   key %s and val %s same in both \n", key1, val1);
#endif

		}
		else if ((pair1 == 0) || (pair2 == 0)) {
			/* one header had a keyword, the other didn't.  Failure. */
#ifdef DEBUG
			printf("    only one key in this pair: %s vs. %s \n",
					key1, key2);
#endif
			return(1);
		}
		else {
			/* 
			 *  neither line was of type "key  = value" 
		 	 *  In this case, just accept as "identical"
			 */
#ifdef DEBUG2
				printf("    accept non-key-val %s vs. %s \n",
					gotit1, gotit2);
#endif	
		}




	}

	return(0);
}




/**********************************************************************
 * PROCEDURE: read_fits_key_val
 *
 * DESCRIPTION: Given a string, which is one line from a FITS header,
 *              try to parse it as a FITS keyword=val line.
 *              This means that the string must have the form
 *
 *                 keyword = value
 *
 *              where
 *         
 *                 "keyword" starts at col 0 and is at most 8 chars long
 *    
 *                 "="       must be in col 8 
 *
 *                 "value"   starts in col 9 and is either
 *                              the first non-blank sequence of chars
 *                           or 
 *                              'stuff inside single quote marks'
 *
 *              If the string does have this format, then write the
 *              keyword and the value into the pre-allocated strings which are
 *              provided.  The user MUST PRE-ALLOCATE SPACE for 
 *              these arguments!
 *      
 * RETURNS:   0        if the line has a valid "keyword = value" format
 *            1        otherwise
 *
 */

static int
read_fits_key_val
	(
	char *header_line,   /* I: the line taken from a FITS header */
	char *keyword,       /* I/O: a pre-allocated string, into which we */
	                     /*        will write the keyword (if found) */
	char *value          /* I/O: a pre-allocated string, into which we */
	                     /*        will write the value (if found) */
	)
{
	int i, j, k, first;
	int quote_flag;
	char err_msg[BIG_BUF], line[CARD_LEN + 1];
	
	/* sanity checks */
	if (keyword == NULL) {
		error(1, "read_fits_key_val: given NULL keyword arg \n");
	}
	if (value == NULL) {
		error(1, "read_fits_key_val: given NULL value arg \n");
	}

	/* create a copy of the first CARD_LEN characters of the line */
   strncpy(line, header_line, CARD_LEN);
	line[CARD_LEN] = '\0';

#ifdef DEBUG
		printf("read_fits_key_val: checking line %s\n", line);
#endif

	/* check to see if char in col 8 is '=' */
	if ( (strlen(line) < 9) || (line[8] != '=') ) {
#ifdef DEBUG
		printf("read_fits_key_val: col 8 doesn't have equals sign \n");
#endif
		return(1);
	}

	/* copy the first 8 characters of the line into the "keyword" arg */
	for (i = 0; i < 8; i++) {
		keyword[i] = line[i];
	}
	keyword[8] = '\0';

	/* 
	 * now, we need to copy the "value" portion of the line. 
	 *
	 * Check to see if the first non-blank char after equals sign
	 *    is a single quote mark '.  If so, then verify that a SECOND
	 *    single quote mark appears before end of the line.
	 *    If there are two quote marks, then copy the quote marks
	 *    and everything in between them into the "value" arg.
	 */
	quote_flag = 0;
	for (i = 9; i < strlen(line); i++) {
		if (i >= CARD_LEN) {
			break;
		}
 		if (line[i] == '\0') {
			break;
		}
		if (line[i] == '\'') {
			for (j = i + 1; j < strlen(line); j++) {
				if (j >= CARD_LEN) {
					break;
				}
				if (line[j] == '\0') {
					break;
				}
				if (line[j] == '\'') {
					quote_flag = 1;
					break;
				}
			}

			if (quote_flag == 1) {
				/* yes, the line contains something inside quote marks. */
				for (k = i; k <= j; k++) {
					value[k - i] = line[k];
				}
				value[k - i] = '\0';
#ifdef DEBUG
				printf("fits_read_key_val: A key ..%s.. val ..%s.. \n", keyword, value);
#endif
				return(0);
			}

		}
	}

	/* if we get here, then the contents of the line after the equals
	 *    sign were NOT surrounded by a pair of single quotes.
	 *    In that case, we simply copy the first word after the
	 *    equals sign into the "value" argument.
	 */

	/* first, skip blanks after the equals sign */
	for (i = 9; i < strlen(line); i++) {
		if (i >= CARD_LEN) {
			break;
		}
 		if (line[i] == '\0') {
			break;
		}
		if (line[i] != ' ') {
			break;
		}
	}
	/* now, copy all subsequent non-blank chars to the "value" arg */
	first = i;
	while ((i < CARD_LEN) && (line[i] != '\0') && (line[i] != ' ')) {
		value[i - first] = line[i];
		i++;
	}
	value[i - first] = '\0';

#ifdef DEBUG
	printf("fits_read_key_val: B key ..%s.. val ..%s.. \n", keyword, value);
#endif

	return(0);
}

