
	/*	
	 * Reads a raw KWFC FITS image, divides all pixel data
	 * by 2 -- so that the data range changes from 0 - 65535
	 * to 0 - 32767 -- and then writes the divided image
	 * out to a new FITS file.
	 *
	 * The new FITS file should be the same size as the original,
	 * but there is one change in the header:
	 *
	 *      In the original file:
	 * 
	 * BZERO   =        32768.0 / offset data range to that of unsigned short
	 *
	 *      In the new file:
	 *
	 * BZERO   =            0.0 / Reset by div2
	 *
	 * One can use this keyword to verify that an image has been
	 * divided by two (or just look at the data values, of course).
	 *	
	 *    
	 *    Usage:   div2   infile  outfile=new_name
	 *
	 * where
	 *                 infile          is original raw KWFC FITS image
	 *  
	 *                 new_name        is name for new, divided-by-2 image
	 *        
	 *
	 * MWR 10/24/2011
	 */



#include "pcvista.h"
#include "fits.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

#undef DEBUG
#undef DEBUG2
#define MAX_HEADER_LINES 1000
#define CARD_LENGTH      80
#define CARDS_PER_BUNCH  36



#define USAGE  "usage: div2  file outfile=  [help=] "


/* 
 * global variables, used by both
 *     check_input_header()
 *     read_and_convert()
 */
char header_array[MAX_HEADER_LINES + 1][CARD_LENGTH + 1];


static int check_input_header(
	char *filename,        /* I: name of FITS file  */
	int *naxis1,           /* O: value of NAXIS1 from header */
	int *naxis2,           /* O: value of NAXIS2 from header */
	double *bscale,        /* O: value of BSCALE from header */
	double *bzero,         /* O: value of BZERO from header */
	int *num_cards         /* O: number of non-blank FITS header cards */
	);

static int
read_and_convert
	(
	char *infile,          /* I: name of input FITS file  */
	char *outfile,         /* I: name of output FITS file (we create it) */
	int naxis1,            /* I: value of NAXIS1 from header */
	int naxis2,            /* I: value of NAXIS2 from header */
	double bscale,         /* I: value of BSCALE from header */
	double bzero,          /* I: value of BZERO from header */
	int num_cards          /* I: number of non-empty FITS header cards */
	);

static int
decode_pixel_value
	(
	unsigned char low_byte,         /* I: low byte of pixel data */
	unsigned char high_byte,        /* I: high byte of pixel data */
	double bzero                    /* I: BZERO from FITS header */
	);

static int
encode_pixel_value
	(
	int pixel_value,                /* I: encode this value */
	unsigned char *low_byte,        /* O: low byte of encoded pixel data */
	unsigned char *high_byte        /* O: high byte of encoded pixel data */
	);




int
main(argc,argv)
int argc;
char *argv[];
{
	char infile[BIG_BUF], outfile[BIG_BUF];
	char *gotit;
	int i;
	int naxis1, naxis2, num_cards;
	double bzero, bscale;
	

	if (argc == 1) {
		fprintf(stderr, "%s\n", USAGE);
		exit(1);
	}
	if (strcmp(argv[1], "help") == 0) {
		printf("%s\n", USAGE);
		exit(0);
	}

	/* prevent string overruns */
	infile[BIG_BUF - 1] = '\0';
	outfile[BIG_BUF - 1] = '\0';
	
	/* initialize stuff */
	strcpy(infile, "");
	strcpy(outfile, "");
	naxis1 = -1;
	naxis2 = -1;
	bscale = 1.0;
	bzero = 0.0;
	
	/* the first argument will be the input file name */
	strncpy(infile, argv[1], BIG_BUF-1);

	for (i = 2; i < argc; i++) {
		if (keyword("help", argv[i])) {
			printf("%s\n", USAGE);
			exit(0);
		}
		if ((gotit = find("outfile", argv[i])) != NULL) {
			strncpy(outfile, gotit, BIG_BUF-1);
			continue;
		}
	}
#ifdef DEBUG
	printf(" infile %s  outfile %s \n", infile, outfile);
#endif
	
	/* check arguments for validity before continuing */
	if ((strlen(infile) == 0) || (strlen(outfile) == 0)) {
		fprintf(stderr, "invalid input and/or output file names \n");
		exit(1);
	}


	/* read the input file header */
	if (check_input_header(infile, &naxis1, &naxis2,
				&bscale, &bzero, &num_cards) != 0) {
		/* an error occurred */
		fprintf(stderr, "div2: check_input_header fails \n");
		exit(1);
	}
	else {
#ifdef DEBUG
		printf(" result of check_input_header: \n");
		printf(" naxis1 %5d  naxis2 %5d \n", naxis1, naxis2);
		printf(" bscale %9.4f  bzero %10.4f \n", bscale, bzero);
		printf(" num_cards %5d \n", num_cards);
#endif
	}

	/* read the input file data, convert and write into output file */
	if (read_and_convert(infile, outfile, naxis1, naxis2,
			bscale, bzero, num_cards) != 0) {
		fprintf(stderr, "div2: read_and_convert fails \n");
		exit(1);
	}
	else {
#ifdef DEBUG
		printf("read_and_convert succeeds \n");
#endif
	}


	exit(0);
}


	/*************************************************************
	 * PROCEDURE: check_input_header
	 *
	 * DESCRIPTION: read the FITS header of the input file.
	 *              Verify that
	 *
	 *                   SIMPLE = T
	 *                   BITPIX = 16
	 *                   NAXIS  = 2
	 * 
	 *              Read the values of NAXIS1 and NAXIS2.
	 *
	 *              Look for BSCALE and BZERO; if found, 
	 *              read their values.
	 *
	 *              Place results into output arguments.
	 *
	 * RETURNS:     0       if all goes well
	 *              1       if an error occurs
	 */


static int
check_input_header
	(
	char *filename,        /* I: name of FITS file  */
	int *naxis1,           /* O: value of NAXIS1 from header */
	int *naxis2,           /* O: value of NAXIS2 from header */
	double *bscale,        /* O: value of BSCALE from header */
	double *bzero,         /* O: value of BZERO from header */
	int *num_cards         /* O: number of FITS header cards */
	                       /*     up to the "END" card */
	)
{
	int i;
	int done;
	int start_card, end_card, card_index;
	int naxis1_val, naxis2_val;
	int simple_flag, naxis_flag, naxis1_flag, naxis2_flag;
	int bscale_flag, bzero_flag;
	double bscale_val, bzero_val;
	FILE *fp;
	


	/* 
	 * set all flags to zero -- signalling no information yet.
	 * As we read bit of info from the header, we check for valid value.
	 * If value is OK, we set flag to 1.
	 */
	simple_flag = 0;
	naxis_flag = 0;
	naxis1_flag = 0;
	naxis2_flag = 0;

	/* 
	 * but since these two values have defaults, set flag to 1.
	 * We don't need to read them from header, we can use defaults.
	 */
	bscale_val = 1;
	bzero_val = 0;
	bscale_flag = 1;
	bzero_flag = 1;

	/* open file for reading */
	if ((fp = fopen(filename, "r")) == NULL) {
		fprintf(stderr, "check_input_header: can't open file %s \n",
				filename);
		return(1);
	}
		

	/* read in all header lines until we find an "END   " line */
	done = 0;
	start_card = 0;
	while (!done) {

#ifdef DEBUG
		printf("starting to read card %d \n", start_card);
#endif
		/* read in a "bunch" of header lines */
		for (i = 0; i < CARDS_PER_BUNCH; i++) {
			card_index = i + start_card;
			if (fread(header_array[card_index], CARD_LENGTH, 1, fp) != 1) {
				fprintf(stderr, "check_input_header: failed to read card %d \n", 
						card_index);
				return(1);
			}
			/* add a string terminator to end of card, for convenience */
			header_array[card_index][CARD_LENGTH] = '\0';
#ifdef DEBUG
			printf(" card %d is ..%s..\n", card_index, header_array[card_index]);
#endif

			/* check to see if we found the END card */
			if (strncmp(header_array[card_index], "END     ", 8) == 0) {
				end_card = card_index;
				done = 1;
			}

		}
		start_card += CARDS_PER_BUNCH;
	}
	fclose(fp);



	/* 
	 * If we get here, then we did succeed in reading the FITS header.
	 *
	 * Now, verify that this is a very simple, BITPIX=16 FITS image,
	 * and get the number of rows and cols.  Check for the 
	 * scaling factors as well.
	 */
	for (i = 0; i <= end_card; i++) {
		char header_line_copy[CARD_LENGTH + 1];
		char *delim, *tok;
		delim = " ";

		/* make a copy of each line so strtok doesn't modify header lines */
		strncpy(header_line_copy, header_array[i], CARD_LENGTH);
		header_line_copy[CARD_LENGTH] = '\0';

		/* verify that SIMPLE = T */
		if (strncmp(header_line_copy, "SIMPLE  =", 9) == 0) {
			tok = strtok(header_line_copy + 9, delim);
#ifdef DEBUG
			printf("tok for SIMPLE is ..%s.. \n", tok);
#endif
			if (strcmp(tok, "T") != 0) {
				fprintf(stderr, "SIMPLE is not T \n");
				return(1);
			}
			simple_flag = 1;
			continue;
		}

		/* verify that NAXIS = 2 */
		if (strncmp(header_line_copy, "NAXIS   =", 9) == 0) {
			tok = strtok(header_line_copy + 9, delim);
#ifdef DEBUG
			printf("tok for NAXIS is ..%s.. \n", tok);
#endif
			if (strcmp(tok, "2") != 0) {
				fprintf(stderr, "NAXIS is not 2 \n");
				return(1);
			}
			naxis_flag = 1;
			continue;
		}

		/* get the value for  NAXIS1 */
		if (strncmp(header_line_copy, "NAXIS1  =", 9) == 0) {
			tok = strtok(header_line_copy + 9, delim);
#ifdef DEBUG
			printf("tok for NAXIS1 is ..%s.. \n", tok);
#endif
			naxis1_val = atoi(tok);
#ifdef DEBUG
			printf("naxis1_val is %d \n", naxis1_val);
#endif
			if ((naxis1_val <= 0) || (naxis1_val >= NMAX)) {
				fprintf(stderr, "invalid value for NAXIS1 %s \n", tok);
				return(1);
			}
			naxis1_flag = 1;
			continue;
		}

		/* get the value for  NAXIS2 */
		if (strncmp(header_line_copy, "NAXIS2  =", 9) == 0) {
			tok = strtok(header_line_copy + 9, delim);
#ifdef DEBUG
			printf("tok for NAXIS2 is ..%s.. \n", tok);
#endif
			naxis2_val = atoi(tok);
#ifdef DEBUG
			printf("naxis2_val is %d \n", naxis2_val);
#endif
			if ((naxis2_val <= 0) || (naxis2_val >= NMAX)) {
				fprintf(stderr, "invalid value for NAXIS2 %s \n", tok);
				return(1);
			}
			naxis2_flag = 1;
			continue;
		}

		/* get the value for  BSCALE (if it exists) */
		if (strncmp(header_line_copy, "BSCALE  =", 9) == 0) {
			tok = strtok(header_line_copy + 9, delim);
#ifdef DEBUG
			printf("tok for BSCALE is ..%s.. \n", tok);
#endif
			bscale_val = atof(tok);
#ifdef DEBUG
			printf("bscale_val is %lf \n", bscale_val);
#endif
			if (bscale_val <= 0) {
				fprintf(stderr, "invalid value for BSCALE %s \n", tok);
				return(1);
			}
			bscale_flag = 1;
			continue;
		}

		/* get the value for  BZERO (if it exists) */
		if (strncmp(header_line_copy, "BZERO   =", 9) == 0) {
			tok = strtok(header_line_copy + 9, delim);
#ifdef DEBUG
			printf("tok for BZERO is ..%s.. \n", tok);
#endif
			bzero_val = atof(tok);
#ifdef DEBUG
			printf("bzero_val is %lf \n", bzero_val);
#endif
			bzero_flag = 1;
			continue;
		}

	}

	/* make sure that
	 *
	 *   a) we read all required information from header
	 *   b) all information we read was valid
	 */
	if ((simple_flag == 0) || (naxis_flag == 0) || 
	    (naxis1_flag == 0) || (naxis2_flag == 0) ||
	    (bscale_flag == 0) || (bzero_flag == 0)) {
		fprintf(stderr, "check_input_header: invalid information in header \n");
		return(1);
	}


	/* set the output args to the values we read from header */
	*naxis1 = naxis1_val;
	*naxis2 = naxis2_val;
	*bscale = bscale_val;
	*bzero  = bzero_val;
	*num_cards = end_card;

	return(0);
}



	/*************************************************************
	 * PROCEDURE: read_and_convert
	 *
	 * DESCRIPTION: We know the parameters of the input FITS image.
	 *              So, we can now
	 *               
	 *                  read the data section of the input image
	 *                  divide all data values by 2
	 *                  write the new values to a new output image
	 *
	 *              We need to copy the FITS header from the input
	 *              FITS image, too.
	 *
	 * RETURNS:     0       if all goes well
	 *              1       if an error occurs
	 */


static int
read_and_convert
	(
	char *infile,          /* I: name of input FITS file  */
	char *outfile,         /* I: name of output FITS file (we create it) */
	int naxis1,            /* I: value of NAXIS1 from header */
	int naxis2,            /* I: value of NAXIS2 from header */
	double bscale,         /* I: value of BSCALE from header */
	double bzero,          /* I: value of BZERO from header */
	int num_cards          /* I: number of non-empty FITS header cards */
	)
{
	int i;
	int num_cards_written;
	FILE *infp, *outfp;

	/* open the input file */
	if ((infp = fopen(infile, "r")) == NULL) {
		fprintf(stderr, "read_and_convert: can't open input file %s \n",
				infile);
		return(1);
	}	

	/* create and open the output file */
	if ((outfp = fopen(outfile, "w")) == NULL) {
		fprintf(stderr, "read_and_convert: can't open output file %s \n",
				outfile);
		return(1);
	}	

	/* copy the FITS header */
			/* this doesn't add a COMMENT line to end ... yet */
	for (i = 0; i <= num_cards; i++) {

		/* If this is BZERO card, we change the value to 0.0 */
		if (strncmp(header_array[i], "BZERO   =", 9) == 0) {
			int j;
			char new_line[CARD_LENGTH + 1];

			for (j = 0; j < CARD_LENGTH; j++) {
				new_line[j] = ' ';
			}
			strcpy(new_line, "BZERO   =  0.0                 / Reset by div2");
			for (j = 0; j < CARD_LENGTH; j++) {
				if (new_line[j] == '\0') {
					new_line[j] = ' ';
				}
			}
			if (fwrite(new_line, CARD_LENGTH, 1, outfp) != 1) {
				fprintf(stderr, "read_and_convert: failed to write new BZERO card \n");
				return(1);
			}
		}
		else {
			if (fwrite(header_array[i], CARD_LENGTH, 1, outfp) != 1) {
				fprintf(stderr, "read_and_convert: failed to write card %d \n", 
						i);
				return(1);
			}
		}
	}
	/* 
	 * now we write additional blank cards to bring the total number of
	 * cards up to a multiple of CARDS_PER_BUNCH
	 */
	num_cards_written = num_cards + 1;
	if (num_cards_written % CARDS_PER_BUNCH != 0) {
		int j;
		char blank_line[CARD_LENGTH];

		for (j = 0; j < CARD_LENGTH; j++) {
			blank_line[j] = ' ';
		}

		i = num_cards;
		while ((i + 1) % CARDS_PER_BUNCH != 0) {
#ifdef DEBUG
			printf(" writing blank line, card %d \n", i);
#endif
			if (fwrite(blank_line, CARD_LENGTH, 1, outfp) != 1) {
				fprintf(stderr, "read_and_convert: failed to write blank %d \n", 
						i);
				return(1);
			}
			i++;
			num_cards_written++;
		}
	}
#ifdef DEBUG
	printf("num_cards_written is %d \n", num_cards_written);
#endif
			
			

	/* copy the FITS data, dividing by 2 as we go */
	{
		int row, col;
		char data_array[2*NMAX], div_data_array[2*NMAX];
		long offset;
		size_t bytes_per_pixel, num_cols;
	
		/* seek to start of data in input file */
		offset = num_cards_written*CARD_LENGTH;
#ifdef DEBUG
		printf("seeking to position %ld in %s \n", offset, infile);
#endif
		if (fseek(infp, offset, SEEK_SET) != 0) {
			fprintf(stderr, "read_and_convert: fseek to %ld in %s fails \n",
						offset, infile);
			return(1);
		}
	
		/* loop over rows: read, convert, write */
		bytes_per_pixel = 2;
		num_cols = naxis1;
#ifdef DEBUG
		printf(" num_cols is %d \n", (int) num_cols);
#endif
		for (row = 0; row < naxis2; row++) {

			if (fread(data_array, bytes_per_pixel, num_cols, infp) != num_cols) {
				fprintf(stderr, "read_and_convert: fread fails row %d \n", row);
				return(1);
			}

	
			for (col = 0; col < num_cols; col++) {
				int original_value, new_value;

				original_value = decode_pixel_value(
							(unsigned char) data_array[2*col + 1], 
							(unsigned char) data_array[2*col], bzero);
#ifdef DEBUG
				if ((row >= 45) && (row <= 52) && (col >= 55) && (col <= 64)) {
					printf(" %5d %5d decode_pixel_value returns %6d \n", 
							row, col, original_value);
				}
#endif

				new_value = original_value / 2;
				if (encode_pixel_value(new_value,
							(unsigned char *) &(div_data_array[2*col + 1]),
							(unsigned char *) &(div_data_array[2*col])) != 0) {
					fprintf(stderr, "read_and_convert: encode_pixel_value fails on row %d col %d \n", row, col);
					return(1);
				}
#ifdef DEBUG
				if ((row >= 45) && (row <= 52) && (col >= 55) && (col <= 64)) {
					printf(" %5d %5d encode_pixel_value returns %5u %5u \n", 
							row, col, 
							(unsigned char) div_data_array[2*col + 1],
							(unsigned char) div_data_array[2*col]);
				}
#endif
			}

			/* now write the divided-by-2 data into output file */
			if (fwrite(div_data_array, bytes_per_pixel, num_cols, outfp) != num_cols) {
				fprintf(stderr, "read_and_convert: fwrite fails row %d \n", row);
				return(1);
			}


		}

	}

	/* 
	 * we must now pad the output file with NUL characters until we
	 * reach a multiple of CARD_LENGTH * CARDS_PER_BUNCH bytes total 
	 */
	{
		char pad_array[CARD_LENGTH*CARDS_PER_BUNCH + 1];
		int i;
		long total_bytes, pad_bytes, bunch_bytes;

		/* file must contain a multiple of this many bytes */
		bunch_bytes = CARD_LENGTH * CARDS_PER_BUNCH;
		total_bytes = ftell(outfp);
		pad_bytes = (((total_bytes/bunch_bytes) + 1)*bunch_bytes) - total_bytes;
#ifdef DEBUG
		printf("  after data, output file has %ld bytes \n", total_bytes);
		printf("  so add %ld bytes to pad it \n", pad_bytes);
#endif
		for (i = 0; i < pad_bytes; i++) {
			pad_array[i] = 0;
		}
		if (fwrite(pad_array, 1, pad_bytes, outfp) != pad_bytes) {
			fprintf(stderr, "read_and_convert: fwrite fails on padding\n");
			return(1);
		}
	}

				



	/* close files */
	fclose(infp);
	fclose(outfp);

	return(0);
}



	/*************************************************************
	 * PROCEDURE: decode_pixel_value
	 *
	 * DESCRIPTION: Given two bytes containing a single pixel
	 *              value from a FITS 16-bit integer file,
	 *              and the BZERO value from FITS header,
	 *              figure out the pixel value.  
	 *
	 *              Algorithm is:
	 *
	 *                  low-byte:   interpret all 8 bits as unsigned int
	 *                                   call this "A"
	 * 
 	 *                  high-byte:  interpret low 7 bits as unsigned int
	 *                                   call this "B"
	 *
	 *                              interpret most significant bit as
	 *                                   -32768    if bit = on
	 *                                        0    if bit = off
	 *                                   call this "C"
	 *
	 *              The final pixel value will be
	 *
	 *                  value =  (A + B + C) + bzero
	 *
	 * RETURNS:     pixel value
	 */


static int
decode_pixel_value
	(
	unsigned char low_byte,         /* I: low byte of pixel data */
	unsigned char high_byte,        /* I: high byte of pixel data */
	double bzero                    /* I: BZERO value from FITS header */
	)
{
	int low_val, high_val;
	int radix_val, pixel_val;

	low_val = (unsigned int) low_byte;
	high_val = (unsigned int) (high_byte & 0x7F);
	if (high_byte & 0x80) {
		radix_val = -32768;
	}
	else {
		radix_val = 0;
	}

	pixel_val = radix_val + high_val*256 + low_val;
	pixel_val += bzero;
#ifdef DEBUG2
	printf(" decode: low %5d  high %5d -> %5d   radix %6d   tot %6d \n",
				low_val, high_val, high_val*256, radix_val, pixel_val);
#endif

	return(pixel_val);
}


	/*************************************************************
	 * PROCEDURE: encode_pixel_value
	 *
	 * DESCRIPTION: Given a pixel value -- which must lie in the
	 *              range from 0 to +32767, since we have 
	 *              divided the original pixel value by 2 --
	 *              encode it into two bytes which can be written
	 *              into a 16-bit integer FITS file.
	 *
	 *              Algorithm is simple, since we know that the 
	 *              pixel value lies in range 0 to +32767.
	 *
	 *                  low_byte   = (value & 0x00FF)
	 * 
 	 *                  high-byte  = (value & 0x7F) 
	 *
	 *
	 *              Place the two bytes into the output arguments.
	 *
	 * RETURNS:     
	 *              0        if all goes well
	 *             -1        if an error occurs
	 */


static int
encode_pixel_value
	(
	int pixel_value,                /* I: encode this value */
	unsigned char *low_byte,        /* O: low byte of encoded pixel data */
	unsigned char *high_byte        /* O: high byte of encoded pixel data */
	)
{
	unsigned char low, high;

	/* sanity check */
	if ((pixel_value < 0) || (pixel_value > 32767)) {
		fprintf(stderr, "encode_pixel_value: invalid pixel value %d \n", 
					pixel_value);
		return(1);
	}

	low = (unsigned char) (pixel_value &  0x00FF);
	high = (unsigned char) ((pixel_value &  0x7F00) >> 8);
#ifdef DEBUG2
	printf(" encode: value  %6d  low %5d high %5d   \n",
				pixel_value, low, high);
#endif

	*low_byte = low;
	*high_byte = high;

	return(0);
}
