
/**************************************************************************
 *                                                                        *
 * Copyright (c) 1996 Michael Richmond and Richard Treffers               *
 *                                                                        *
 *                    This software may be copied and distributed for     *
 *                    educational, research and not for profit services   *
 *                    provided that this copyright and statement are      *
 *                    included in all such copies.                        *
 *                                                                        *
 **************************************************************************/


/*	
	Collapses data to generate spectra

	  MASH file SP=i1,i2 [SP=i3,i4] [BK=i6,i7] [NORM] [COLS] [ROWS]

	makes columns of spectrum and background

	9/22/92 - output sent to stdout and not .spc
			- also changed usage, exit(0)
			- the upper limit is NOT included  -rrt
	7/22/93 - add box -rrt


	Changed the way the "sky" value is calculated using the 
	regions specified by the "sk=" rows (or columns).  
	Old method required the size of this region to be exactly
	the same as the size of the spectral region, and used a 
	simple sum (or average) of all pixel values.  New method
	allows a sky region of arbitrary size, and uses a clipped
	mean method.
	
	MWR 12/28/2017

*/

#undef DEBUG
#undef DEBUG2

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include "pcvista.h"
#include "fits.h"

#ifdef PROTO
int main(int, char **);
static void mash_rows(int ,int, int, double);
static void mash_cols(int, int, int, double);
static int calc_clipped_mean(int num_pix, int16 *data[], int num_iter,
                             double nsigma, double *clipped_mean);
#else
static void mash_rows();
static void mash_cols();
static int calc_clipped_mean();
#endif

/* max number of background segments which can be defined */
#define NBIN 5
/* 
 * default value for background determination: number of iterations 
 * of sigma-clipping calculations */
#define BACK_ITER 3
/* 
 * default value for background determination: ignore values 
 * which are more than this many times the stdev away from mean 
 */
#define BACK_NSIGMA 2

static int mean_flag;
static int col_flag;
static int nspect, nback;
static int nspect_total, nback_total;
static int slower[NBIN], supper[NBIN], blower[NBIN], bupper[NBIN];
static FITS_HANDLE fh;
static int16 data[NMAX];
static int nrow, ncol;



#define USAGE \
  "usage: mash file sp=i1,i2 [sp=i3,i4] [bk=i6,i7] [norm] [cols] [rows] \n      [box=] [back_iter=] [back_nsigma=] [help]"

int
main(argc, argv)
int argc;
char *argv[];
{
	int i, nfield, boxno=-1;
	int back_iter = BACK_ITER;
	char *gotit, fname[NBUF], string1[NBUF], string2[NBUF];
	int sr,sc,nr,nc;
	double back_nsigma = BACK_NSIGMA;

	if (argc == 1) {
		error(-1, USAGE);
	}
	if (strcmp(argv[1], "help") == 0) {
		printf("%s\n", USAGE);
		exit(0);
	}

	for (i = 1; i < argc; i++) {
		if ((gotit = find("sp", argv[i])) != NULL) {
			if (nspect >= NBIN)
				error(-1, "too many spectrum sections to handle");
			nfield = sscanf(gotit,"%[^,],%s", string1, string2);
			if (nfield != 2)
				error(-1, "syntax SP=i1,i2");
			slower[nspect] = (int) (evaluate(string1) + 0.5);
			supper[nspect] = (int) (evaluate(string2) + 0.5);
			nspect_total += (supper[nspect] - slower[nspect]);
			nspect++;
			continue;
		}
		if ((gotit = find("bk", argv[i])) != NULL) {
			if (nback >= NBIN)
				error(-1, "too many background sections to handle");
			nfield = sscanf(gotit,"%[^,],%s", string1, string2);
			if (nfield != 2)
				error(-1, "syntax BK=i1,i2");
			blower[nback] = (int) (evaluate(string1) + 0.5);
			bupper[nback] = (int) (evaluate(string2) + 0.5);
			nback_total += (bupper[nback] - blower[nback] );
			nback++;
			continue;
		}
		if (keyword("norm", argv[i])) {
			mean_flag = 1;
			continue;
		}
		if (keyword("rows", argv[i])) {
			col_flag = 0;
			continue;
		}
		if (keyword("cols", argv[i])) {
			col_flag = 1;
			continue;
		}
		if ((gotit = find("box", argv[i])) != NULL) {
			boxno = (int)evaluate(gotit);
			continue;
		}
		if ((gotit = find("back_iter", argv[i])) != NULL) {
			back_iter = (int)evaluate(gotit);
			continue;
		}
		if ((gotit = find("back_nsigma", argv[i])) != NULL) {
			back_nsigma = (double)evaluate(gotit);
			continue;
		}
		strcpy(fname,argv[i]);		/* generate output spectrum file name*/
	}

	fh = fits_open(fname, "r", &nrow, &ncol);
	if(boxno != -1){
		if (getbox(boxno, &sr, &sc, &nr, &nc))
				error(-1, "can't find box");
		check_box(boxno, nrow, ncol);
	}else{
		sr=0;
		sc=0;
		nr=nrow;
		nc=ncol;
	}
	if (nspect_total == 0)		/* correct for number of rows */
		error(-1,"no rows selected ");


	if (col_flag) {
		mash_cols(sr,nr, back_iter, back_nsigma);
	}
	else {
		mash_rows(sc,nc, back_iter, back_nsigma);
	}

	exit(0);
}

static void
mash_cols(sr,nr, num_iter, nsigma)
int sr,nr;
int num_iter;
double nsigma;
{
	int i, j,row, ibin, num_back;
	double spect_sum, sky_val;
	int16 back_pixels[NBIN*NMAX];
	double old_sky_sum, new_sky_sum;

	for (row = sr , j=0 ; j < nr ; j++, row++) {
		fits_get_data(fh, row, 0, data, ncol);

		/* add up the (signal+background) from the spectral region */
		spect_sum = 0.0;
		for (ibin = 0; ibin < nspect; ibin++) {
			for (i = slower[ibin]; i < supper[ibin]; i++) {
				spect_sum += (double) data[i];
			}
		}


		/* compute the sky contribution the old way, just for reference */
		old_sky_sum = 0.0;
      for (ibin = 0; ibin < nback; ibin++) {
         for (i = blower[ibin]; i < bupper[ibin]; i++) {
            old_sky_sum += (double) data[i];
			}
		}


		/* now figure out the clipped mean in the sky region */
		num_back = 0;
		for (ibin = 0; ibin < nback; ibin++) {
			for (i = blower[ibin]; i < bupper[ibin]; i++) {
#ifdef DEBUG2
				printf("    data [ %5d ] = back_pixel %5d = %5d \n",
						i, num_back, data[i]);
#endif
				back_pixels[num_back++] = data[i];
			}
		}
		if (calc_clipped_mean(num_back, back_pixels, num_iter,
							nsigma, &sky_val) != 0) {
			error(1, "mash: calc_clipped_mean failed in mash_cols");
		}
		/* subtract the sky contribution */
		new_sky_sum = (double) (sky_val* (double) nspect_total);
#ifdef DEBUG
		printf(" old_sky_sum %8.2f   new_sky_sum = %8.2f * %5d = %8.2f \n",
				old_sky_sum, sky_val, nspect_total, new_sky_sum);
#endif
		spect_sum -= (double) new_sky_sum;

		/* if desired, normalize the result */
		if (mean_flag)
			spect_sum /= nspect_total;
		printf("%d %8.2f\n", row, spect_sum);
	}
}

static void
mash_rows(sc,nc, num_iter, nsigma)
int sc,nc, num_iter;
double nsigma;
{
	int i, row, icol, ibin;
	int num_back[NMAX];
	int16 *back_pixels[NMAX];
	double sky_val;
	static double spect_sum[NMAX];
	static double back_sum[NMAX];
	static double old_sky_sum[NMAX], new_sky_sum[NMAX];

	/* allocate memory for the array of background pixel values */
	for (i = 0; i < nc; i++) {
		if ((back_pixels[i] = (int16 *) calloc(NMAX, sizeof(int16))) == NULL) {
			error(1, "mash_rows: calloc for back_pixels fails");
		}
	}
		
	/* initialize the spect_sum[] array values to zero */
	for (i = 0; i < nc; i++) {
		spect_sum[i] = 0.0;
	}
	/* initialize the count of the number of background pixels to zero */
	for (icol = 0; icol < nc; icol++) {
		num_back[icol] = 0;
	}
			

	for (ibin = 0; ibin < nspect; ibin++) {
		for (row = slower[ibin] ; row < supper[ibin]; row++) {
			fits_get_data(fh, row, 0, data, ncol);

			/* add up the (signal+background) from the spectral region */
			for (i = 0; i < nc; i++) {
				spect_sum[i] += (double) data[i+sc];
			}
		}
	}

	/* compute the sky value the old way, just for reference */
	for (i = 0; i < nc; i++) {
		back_sum[i] = 0;
	}
   for (ibin = 0; ibin < nback; ibin++) {
      for (row = blower[ibin]; row < bupper[ibin]; row++) {
         fits_get_data(fh, row, 0, data, ncol);
         for (i = 0; i < nc; i++)
            back_sum[i] += (double) data[i+sc];
      }
   }


	/* fill up arrays with background pixel vals, for use in a moment */
	for (ibin = 0; ibin < nback; ibin++) {
		for (row = blower[ibin]; row < bupper[ibin]; row++) {
			fits_get_data(fh, row, 0, data, ncol);
			for (i = 0; i < nc; i++) {
				back_pixels[i][num_back[i]++] = (double) data[i + sc];
			}		
		}
	}
	/* 
	 * determine the background value for each row, and then
	 *    subtract that background value (multiplied by the number of
	 *    pix in spectral region) from the spectral sum for each row
	 */
	for (i = 0; i < nc; i++) {
		if (calc_clipped_mean(num_back[i], back_pixels[i], num_iter,
							nsigma, &sky_val) != 0) {
			error(1, "mash: calc_clipped_mean failed in mash_rows");
		}
		/* subtract the sky contribution */
		new_sky_sum[i] = (double) (sky_val* (double) nspect_total);
#ifdef DEBUG
		printf(" old_sky_sum %8.2f   new_sky_sum = %8.2f * %5d = %8.2f \n",
				old_sky_sum[i], sky_val, nspect_total, new_sky_sum[i]);
#endif
		spect_sum[i] -= (double) new_sky_sum[i];

		/* normalize the result, if desired */
		if (mean_flag) {
			spect_sum[i] /= (double) nspect_total;
		}

		/* and print out the results */
		printf("%d %8.2f\n", i+sc, spect_sum[i]);

	}

	/* de-allocate the arrays used to store background pixel values */
	for (i = 0; i < nc; i++) {
		free(back_pixels[i]);
	}

}






/****************************************************************************
 * PROCEDURE: calc_clipped_mean
 *
 * DESCRIPTION: Given an array of 'npix' pixel values, and a pair 
 *              of parameters 'num_iter' and 'nsigma', compute a
 *              clipped mean of the values.  Discard all values
 *              more than 'nsigma' away from the mean, and re-compute
 *              'num_iter' times.  
 *
 *              In other words, if 'num_iter' = 1, then we perform
 *
 *                   A.  compute mean and stdev of all pixels
 *
 *                   B.  set (min, max) boundaries for acceptable values
 *
 *                   C.  re-compute mean and stdev of pixels between
 *                              (min, max) only
 *
 *              Place the calculated clipped mean into the output
 *              argument 'clipped_mean'.
 *
 * RETURNS:
 *    0          if all goes well
 *    1          if there's a problem
 */

static int
calc_clipped_mean
	(
	int num_pix,              /* I: number of pixel values in array */
	int16 *pix_data,          /* I: array of pixel values */
	int num_iter,             /* I: how many times to iterate? */
	double nsigma,               /* I: ignore pixels more than this far from mean */
	double *clipped_mean      /* O: result of the calculation */
	)
{
	int i_pix, iter;
	int16 clip_min, clip_max;
	double val, sum, pixnum, mean, sumsq, diff, stdev;

	/* sanity checks */
	if (num_pix <= 0) {
		error(1, "mash: calc_clipped_mean handed num_pix <= 0 ?! \n");
	}
	if (data == (int16 *) NULL) {
		error(1, "mash: calc_clipped_mean handed NULL *data ?! \n");
	}
	if (num_iter < 0) {
		error(1, "mash: calc_clipped_mean handed num_iter < 0 ?! \n");
	}
	if (nsigma <= 0) {
		error(1, "mash: calc_clipped_mean handed nsigma <= 0 ?! \n");
	}

	/* calculate clipped mean, stdev in this square */
	clip_min = (int16) MIN_DATA_VAL;
	clip_max = (int16) MAX_DATA_VAL;

	for (iter = 0; iter < num_iter; iter++) {
#ifdef DEBUG2
			printf("   iter %3d  clip %6d %6d \n", 
								 	iter, clip_min, clip_max);
			printf("     ");
			for (i_pix = 0; i_pix < num_pix; i_pix++) {
				printf(" %5d ", pix_data[i_pix]);
			}
			printf("\n");
#endif
			sum = 0.0;
			sumsq = 0.0;
			pixnum = 0;

			for (i_pix = 0; i_pix < num_pix; i_pix++) {
				val = (double) pix_data[i_pix];
				if ((val >= clip_min) && (val <= clip_max)) {
					sum += val;		
					sumsq += (val*val);
					pixnum += 1.0;
				}
			}

			/* find the mean and stdev for this iteration ... */
			if (pixnum == 0.0) {
				error(1, "mash: no pixels within clipping range?");
			}
			mean = sum/pixnum;
			diff = sumsq/pixnum - mean*mean;
			if (diff >= 0.0)
				stdev = sqrt(diff);
			else {
				error(0, "mash: error in std. dev, using 1.0");	
				stdev = 1.0;
			}
#ifdef DEBUG
			printf("    mean %10.2f  stdev %10.2f \n", mean, stdev);
#endif

			/* and set the min, max clip values for next iteration */
			if (mean - nsigma*stdev > MIN_DATA_VAL) {
				clip_min = (int16) (mean - nsigma*stdev);
			}
			if (mean + nsigma*stdev < MAX_DATA_VAL) {
				clip_max = (int16) (mean + nsigma*stdev);
			}
		}

	*clipped_mean = mean;

	return(0);
}


