/**************************************************************************
 *                                                                        *
 * Copyright (c) 1996 Michael Richmond and Richard Treffers               *
 *                                                                        *
 *                    This software may be copied and distributed for     *
 *                    educational, research and not for profit services   *
 *                    provided that this copyright and statement are      *
 *                    included in all such copies.                        *
 *                                                                        *
 **************************************************************************/


/*	
	MULTIPLIES data files

		MUL filea [fileb] [const=c] [div=] [box=] [help]

	multiplies by constant if const is specified
	scales by div if specified (to get non-integral multipliers)
	if BOX= is used, only multiplies over area of box

	1/29/91 - added 'floor' to expression for "const" and "div", 
	          and included <math.h>  - MWR
	9/9/92  - added check for valid data value before assigning result
	          of calculation (in 32 bits) to 16-bit output value MWR 
	9/15/92 - exit(0) added -rrt
	11/17/92 - added outfile option -rrt
	1/7/96   - added 'help' option.  MWR.

	3/9/2006 - fixed two bugs which occurred if user gave "box=" arg:
	                 error in "for" loop boundaries
	                 error using "sr" instead of "sc" as arg in calls
	                       to "fits_get_data()" and "fits_put_data()"
	7/25/2011 - added "pmask=" option.  MWR

	6/14/2013 - modified so if user specifies "outfile=", we copy
	                entire contents to new output file first,
	                then perform the requested multiplication.  MWR

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <string.h>
#include "pcvista.h"
#include "fits.h"

#ifdef PROTO
int main(int, char **);
#endif

char *usage = "mul file [file2] [const=] [div=] [box=] [outfile=] [pmask] [help]";

int
main(argc, argv)
int argc;
char *argv[];
{
	int i, constant = 1, nrow, ncol, div = 1, other_nrow, other_ncol;
	int row, boxno;
	static int nc, nr, sr, sc;
	static int pmask_nrow, pmask_ncol;
	static int pmask_flag = 0;
	long lresult;
	FITS_HANDLE fh, fother = 0,fout, pmask_fh;
	static int16 data[NMAX], other[NMAX], pmask_data[NMAX];
	char *gotit,outfile[NBUF], pmask_file[NBUF];
	char err_msg[NBUF];
	int new_file=0;

	if (argc == 1) { 
		error(-1, usage);
	}
	if (strcmp(argv[1], "help") == 0) {
		printf("%s\n", usage);
		exit(0);
	}

	fh = fits_open(argv[1], "r", &nrow, &ncol);
	nr = nrow; 	
	nc = ncol;

	strcpy(outfile,argv[1]);

	for (i = 2; i < argc; i++) {
		if (keyword("help", argv[i])) {
			printf("%s\n", usage);
			exit(0);
		}
		if ((gotit = find("box", argv[i])) != NULL) {
			boxno = (int) (evaluate(gotit) + 0.5);
			if (getbox(boxno, &sr, &sc, &nr, &nc))
				error(-1, "can't find box");
			check_box(boxno, nrow, ncol);			
			continue;
		}
		if ((gotit = find("const", argv[i])) != NULL) {
			constant = (int) floor(evaluate(gotit) + 0.5);
			continue;
		}
		if ((gotit = find("outfile", argv[i])) != NULL) {
			strcpy(outfile,gotit);
			new_file=1;
			continue;
		}
		if ((gotit = find("pmask", argv[i])) != NULL) {
			strcpy(pmask_file,gotit);
			pmask_flag = 1;
			continue;
		}
		if ((gotit = find("div", argv[i])) != NULL) {
			div = (int) floor(evaluate(gotit) + 0.5);
			if (div == 0)
				error(-1, "can't divide by zero");
			continue;
		}
		if (fother)
			error(-1, "file already open");
		fother = fits_open(argv[i], "r", &other_nrow, &other_ncol);
	}

	if (fother) {
		if (other_nrow != (sr + nr))
			error(-1, "multiplier file does not have same number of rows");
		if (other_ncol != (sc + nc))
			error(-1, "multiplier file does not have same number of columns");
	}

	fout = fits_open(outfile, new_file?"w":"x", &nrow, &ncol);

	if (pmask_flag == 1) {
		pmask_fh = fits_open(pmask_file, "r", &pmask_nrow, &pmask_ncol);
		if (nrow != pmask_nrow) {
			sprintf(err_msg, "number of rows in mask %s != number in %s \n",
					pmask_file, argv[1]);
			error(-1, err_msg);
		}
		if (ncol != pmask_ncol) {
			sprintf(err_msg, "number of cols in mask %s != number in %s \n",
					pmask_file, argv[1]);
			error(-1, err_msg);
		}
	}


	/*
	 * If the user specifies "outfile=", then we do two steps:
	 *   1.  copy original file into new output file, as-is
	 *   2.  then, operate on the new output file as directed by user.
	 * If we don't perform a complete copy first, and user requests
	 * action only within a box, then the regions outside the box
	 * won't have valid data copied into them.
	 */
	if (new_file) {
		fits_copyheader(fh, fout, FITS_CHECK);
		for (row = 0; row < nrow; row++) {
			fits_get_data(fh, row, 0, data, ncol);
			fits_put_data(fout, row, 0, data, ncol);
		}
	}
			
	for (i = 0; i < nc; i++)
		other[i] = 1;

	for (row = sr; row < (sr + nr); row++) {
		fits_get_data(fh, row, sc, data, nc);
		if (pmask_flag == 1) {
			fits_get_data(pmask_fh, row, sc, pmask_data, nc);
		}
		if (fother)
			fits_get_data(fother, row, sc, other, nc);
		for (i = 0; i < ncol; i++) {
			if ((pmask_flag == 0) || (pmask_data[i] == 0)) {
				lresult = (((long)constant*data[i]*other[i])/div);		
			}
			else {
				lresult = data[i];
			}
			if (lresult >= MAX_DATA_VAL)
				data[i] = MAX_DATA_VAL;
			else if (lresult <= MIN_DATA_VAL)
				data[i] = MIN_DATA_VAL;
			else
				data[i] = (int16) lresult;
		}
		fits_put_data(fout, row, sc, data, nc);
	}
	fits_close(fout);
	if (pmask_flag == 1) {
		fits_close(pmask_fh);
	}
	exit(0);
}
