
/**************************************************************************
 *                                                                        *
 * Copyright (c) 1996 Michael Richmond and Richard Treffers               *
 *                                                                        *
 *                    This software may be copied and distributed for     *
 *                    educational, research and not for profit services   *
 *                    provided that this copyright and statement are      *
 *                    included in all such copies.                        *
 *                                                                        *
 **************************************************************************/


/*	
	Inserts a non-linearity into an existing FITS image.

		NONLIN file [outfile=] [MIN=] [MAX=] [POW=] [SCALE=] [box=d]

	Creates a new value for each pixel, calculated like so:

	          if (old <= min)                 no change
	          if (old >= max)                 no change
	          otherwise            
	            new = old + (old - min)*exp(- [(old - min)/scale]^pow  )

	operates over box if specified

	11/20/2008 -- MWR
*/

/*
 * Added "#include <math.h>", needed for "pow()" function.  
 *
 * 3/10/2017 -- MWR
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "pcvista.h"
#include "fits.h"
#include <unistd.h>
#include <string.h>

#ifdef PROTO
int main(int, char **);
#endif

#undef DEBUG

	/* some defaults */
#define DEFAULT_MIN         0.0
#define DEFAULT_MAX      MAX_DATA_VAL
#define DEFAULT_POW         2.0
#define DEFAULT_SCALE    2*MAX_DATA_VAL

#define USAGE \
  "usage: nonlin file [max=] [min=] [pow=] [scale=] [box=] [outfile=] \n       [verbose] [help]"

int verbose;

int
main(argc, argv)
int argc;
char *argv[];
{
	FITS_HANDLE fh,fout;
	int i, boxno;
	static int sr, sc, nr, nc;
	int nrow, ncol, row;
	static int min, max, val;
	double scale, power, new;
	static int16 data[NMAX], clip_hi, clip_lo, new16;
	char *gotit;
	char outfile[NBUF];
	int new_file=0;

	if (argc == 1) {
		error(-1, USAGE);
	}
	if (strcmp(argv[1], "help") == 0) {
		printf("%s\n", USAGE);
		exit(0);
	}

	/* set the default values for command-line parameters */
	min = DEFAULT_MIN;
	max = DEFAULT_MAX;
	scale = DEFAULT_SCALE;
	power = DEFAULT_POW;

	fh = fits_open(argv[1], "r", &nrow, &ncol);
	strcpy(outfile,argv[1]);
	nr = nrow;	
	nc = ncol;

	for (i = 2; i < argc; i++) {
		if (keyword("help", argv[i])) {
			printf("%s\n", USAGE);
			exit(0);
		}
		if ((gotit = find("box", argv[i])) != NULL) {
			boxno = (int)(evaluate(gotit) + 0.5);
			if (getbox(boxno, &sr, &sc, &nr, &nc))
				error(-1, "can't find box");		
			check_box(boxno, nrow, ncol);
			continue;
		}
		if ((gotit = find("min", argv[i])) != NULL) {
			min = (int) evaluate(gotit);
			clip_lo = 1;
			continue;
		}
		if ((gotit = find("max", argv[i])) != NULL) {
			max = (int) evaluate(gotit);
			clip_hi = 1;
			continue;
		}
		if ((gotit = find("pow", argv[i])) != NULL) {
			power = (double) evaluate(gotit);
			continue;
		}
		if ((gotit = find("scale", argv[i])) != NULL) {
			scale = (double) evaluate(gotit);
			continue;
		}
		if ((gotit = find("outfile", argv[i])) != NULL) {
				strcpy(outfile,gotit);
				new_file=1;
				continue;
		}
		if(keyword("verbose",argv[i])){
			verbose=1;
			continue;
		}
	}
	fout = fits_open(outfile, new_file?"w":"x", &nrow, &ncol);
		 
	 if(new_file)
		 fits_copyheader(fh,fout,FITS_CHECK);


	if (verbose) {
		printf("  values between %10d and %10d will be modified to \n",
							min, max);
		printf("       new = %d +  (old - %d)*exp(- [(old - %d)/%9.4e]^%f ) \n",
							min, min, min, scale, power);
	} 

	for (row = sr; row < (sr + nr); row++) {
		fits_get_data(fh, row, sc, data, nc);
		for (i = 0; i < nc; i++) {
			val = data[i];
			if (val <= min) {
				data[i] = val;
#ifdef DEBUG
				printf("  pixel %6d < %6d so leave alone \n", val, min);
#endif
				continue;
			}
			else if (val >= max) {
#ifdef DEBUG
				printf("  pixel %6d > %6d so leave alone \n", val, max);
#endif
				data[i] = val;
				continue;
			}

		{
			double xx, yy, zz, ww;
			xx = (val - min);
			yy = xx/scale;
			zz = pow(yy, power);
			ww = exp(-zz);

			new = min + (val - min)*ww;
#ifdef DEBUG
			printf("  val %6d  xx %9.1f yy %9.1f zz %9.3f ww %9.3f new %9.1f \n",
						val, xx, yy, zz, ww, new);
#endif
		}
			new16 = (int16) new;
			data[i] = new16;
		}
		fits_put_data(fout, row, sc, data, nc);
	}
	fits_close(fout);
	exit(0);
}
