
/**************************************************************************
 *                                                                        *
 * Copyright (c) 1996 Michael Richmond and Richard Treffers               *
 *                                                                        *
 *                    This software may be copied and distributed for     *
 *                    educational, research and not for profit services   *
 *                    provided that this copyright and statement are      *
 *                    included in all such copies.                        *
 *                                                                        *
 **************************************************************************/


/*	include file for PCVISTA
		9/23/92 heavily revised removed ibm*, tec*, pga* screen*,SCREEN*
				removed many globals.
				raised NMAX --> 2048
        7/8/93  NMAX --> 4096 (how times have changed)

	6/14/2013:  Added OLD_CENTER_ALGORITHM option.  By default, not defined.
	               If defined, we use old algorithm to fit gaussian to 
	                   stellar data; will usually yield not-optimal
	                   central positions.
	               if _not_ defined (default), uses a new algorithm
	                   for stellar data, which does allow center 
	                   position to be adjusted from initial guess.
*/



typedef unsigned short uint16;		/* an unsigned 16-bit integer type */
typedef short   int16; 		/* a 16-bit integer type */
typedef int     int32;		/* a 32-bit integer type */

#define	NMAX	16000	/* maximum row size 		*/
#define	NBUF	80
#define BIG_BUF	255
#define NBOX	20		/* maximum number of boxes 	*/
#define NHIST   65536	/* max number of bins in a histogram */
#define SMALL	1.0e-8
#define NWINDOW		16		/* max number of image windows open at once */
                       		/* can be handled by dir-searching routines */

	/* PCVISTA uses 16-bit signed integers to store data */
#define MAX_DATA_VAL  32767		/* maximum data value per pixel */
#define MIN_DATA_VAL -32767		/* minimum data value per pixel */


	/* 
	 * The behavior of the gauss_fit() function was changed in June 2013.
	 * We can describe the changes as follows:
	 *
	 *   - in "old" algorithm, when fitting a gaussian to some distribution,
	 *         we make an initial guess and then iterate, varying the
	 *         three parameters as we go: center, width, amplitude.
	 *         For all three parameters, the stepsize for variations
	 *         is initially set to 10% of the initial guess.
	 *
	 *   - in "new" algorithm, we again make initial guess and then
	 *         iterate.  However, in this scheme, the stepsize for
	 *         _center_ of the distribution depends on what we are
	 *         trying to fit:
	 *            a) if we are fitting to the light of a star or galaxy,
	 *                  use a stepsize of 0.1 (pixels)
	 *            b) if we are fitting to anything else (sky distribution ...),
	 *                  use stepsize of 10% * original guess for center
	 *
	 * Under the "old" algorithm, the stepsize is usually so big that
	 * the iterative process never changes the center position from its
	 * initial guess.  Under the "new" algorithm, the smaller stepsize
	 * does allow the center position to change slightly.
	 * 
	 * By default, we now use the "new" algorithm.
	 * In order to force use of the "old" algorithm in all cases,
	 * set the OLD_CENTER_STEPSIZE variable to 1.   
	 *
	 * MWR 6/13/2013
	 */ 
#define OLD_CENTER_ALGORITHM    0

	/* 
	 * Because calling "gauss_fit()" is so expensive in the star-finding
	 * process, we now add an argument: the maximum number of iterations
	 * to make in the fitting.  This is the default value, and is used
	 * by most routines; but the "stars" program has a command-line
	 * argument to allow the user to change the value, in order
	 * to speed up execution.
	 *
	 * I've found that 20 is a pretty good compromise value; was
	 * the default for many years.
	 *
	 * MWR 6/15/2019
	 */
#define GAUSSFIT_MAXITER  20

	/* values of final arg "pixel_flag" for gauss_fit() function */
#define NO_PIXEL_CENTER 0			/* we are NOT fitting a stellar centroid */
#if OLD_CENTER_ALGORITHM == 1
#define PIXEL_CENTER    NO_PIXEL_CENTER
#else
#define PIXEL_CENTER    1			/* we are fitting a stellar centroid */
#endif


void autospan(int fh,int * span,int *zero,int sr,int sc,int nr,int nc );
int16 **boxdata(int fh, int boxno, int imrows, int imcols, int *boxrows, int *boxcols);
void check_box(int boxno, int nrow, int ncol);
void check_circle(int circno, int nrow, int ncol);
int do_axes(int16 **p, int sr, int sc, int nr, int nc, 
            double sky, double fitrad,
            double *row_cen, double *col_cen, 
            double *row_width, double *col_width,
            double *fwhm, double *peak, double *eccen, 
            double *major_axis, double *minor_axis, int maxter);
int find_sky(int fh, int sr, int sc, int nr, int nc,
            int nskip, int binsize, int min_accept, int max_accept,
            int pmask_flag, int pmask_fh,
				int16 *sky, float *skysig);
void error(int level, char *msg);
void disable_warnings(void);
void enable_warnings(void);
void exten(char *fname,char *ext);
double evaluate(char *);
char *find(char *, char *);
int find_chord(int dr,int rad);
int gauss_fit(float *x, float *y, float *sig, int ndata, float *amp,
              float *center, float *width, float *chisq, int pixel_flag,
              int maxiter);
int getbox(int boxno,int * sr, int * sc, int *nr, int * nc);
int getcircle(int circle,int * cr, int * cc, int * rad);
double get_col(char *buf, int col);
int get_data(int x,int y,int *prow,int *pcol,int16 *pdata,int ndata,char **fname,int* bin,int *zoom);	
int get_win_data(char **fname, 
	int *x_origin, int *x_end, int *y_origin, int *y_end,  
	int *zoom, int *sr, int *endrow, int *sc, int *endcol, int *bin,
	int *zero, int *span, int *dither );
char *getsym(char *);
void image_init(int colormap, int verbose);
void image_quit(void);
int int_box(int *,int *,int *,int *);
int int_circle(int *,int *,int *);
int keyword(char *,char *);
void putsym(char *);
int show_box(int sr, int sc, int er, int ec);
int show_circle(int cr, int cc,  int rad);
void sortem(char **);
void store_window(char *fname,int x_origin,int y_origin,int zoom,int sr,int endrow,int sc,int endcol,int bin, int zero, int span, int dither);
void trim_window(int, int, int, int, int *, int *);
int vista_usleep(int microseconds);
int window_init(void);
int xform_window(int i,int row,int col,int *px,int *py);
int xform_in_window(int i,int row,int col,int *px,int *py);
void	set_win_id(int);
void X_fill(int, int, int, char *);
void X_box(int, int, int, int);
int gauss_matrix(double **matrix, int num, double *vector);
