
/**************************************************************************
 *                                                                        *
 * Copyright (c) 2001 Michael Richmond and Richard Treffers               *
 *                                                                        *
 *                    This software may be copied and distributed for     *
 *                    educational, research and not for profit services   *
 *                    provided that this copyright and statement are      *
 *                    included in all such copies.                        *
 *                                                                        *
 **************************************************************************/


/**************************************************************************
 * Converts image files from the SBIG (Santa Barbara Instruments Corporation)
 * "uncompressed" format into FITS.  This is a very simple procedure,
 * as the SBIG files store a single 2-D, 16-bit integer image.
 *
 * The SBIG format has an ASCII header, with "keyword = value" pairs;
 * these header lines are not fixed at 80 characters, but end as soon
 * as possible.  They use a combination of \n and \r characters to
 * mark the end of each header line.  The "values" in these lines
 * are not quoted with apostrophes, as in FITS.  Thus,
 *
 *     SBIG          Time = 00:45:00 \n\r
 *     FITS          TIME-OBS= '00:45:00'              (to 80 chars)
 *
 * The header of each SBIG image is exactly 2048 characters long.
 * After that, the data portion begins.  Each pixel is represented
 * by two bytes:  the first byte contains the low-order numerical
 * information (i.e. is the low-order byte), and the second byte
 * the high-order informatio (i.e. is the high-order byte).
 * SBIG data values are unsigned.  I choose to convert these to
 * valid 16-bit signed integer values (as required for FITS)
 * by subtracting 32,768.  Thus, a pixel which SBIG software
 * claims is "10" counts will be converted to "-36758" counts
 * in the output file.
 *
 * This program is not designed for general purpose: it exists 
 * to convert images take at the RIT Observatory with our SBIG ST-8
 * camera to a FITS equivalent.
 *
 * Usage:
 *             sbig2fits   sbigfile  fitsfile
 * where
 *        sbigfile        is the name of the SBIG image
 *        fitsfile        is the name of the output FITS image
 *
 *
 * MWR 7/28/2001
 *
 * Modified to take as input SBIG "compressed" images.
 * See the document
 *
 *       http://www.sbig.com/pdffiles/file.format.pdf
 *
 * for information on this format.
 *
 * MWR 6/10/2008
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include "pcvista.h"
#include "fits.h"


#undef DEBUG


/* Starting location of pixel data in SBIG file (bytes from start of file) */
#define SBIG_DATA_START  2048

/*
 * This structure contains all the information we need to create an
 * equivalent FITS file.  We don't transfer all the header info,
 * just the parts I deem important.
 */
#define VAL_LEN  80
typedef struct header_info {
	int nrow;                 /* number of rows in image */
	int ncol;                 /* number of cols in image */
	char dateobs[VAL_LEN];    /* date of observation (presumably UT) */
	char timeobs[VAL_LEN];    /* time of observation (presumably UT) */
	char filter[VAL_LEN];     /* filter  */
	double exptime;           /* exposure time (seconds) */
	                          /*   I believe SBIG records this in hundredths */
	                          /*    of a second, so we must convert */
	double ccd_temp;          /* temperature of CCD (degrees C) */
	double ccd_gain;          /* gain of CCD (electrons per DN) */
} HEADER_INFO;


int main(int, char **);
static int read_header_info(FILE *fin, HEADER_INFO *info);
static int fill_fits_header(FITS_HANDLE handle, HEADER_INFO *info);
static int copy_pixel_data(FILE *fin, FITS_HANDLE handle, HEADER_INFO *info);




#define USAGE "usage: sbigc2fits sbigfile fitsfile  [help]"

int
main(argc, argv)
int argc;
char *argv[];
{
	FILE *fin;
	FITS_HANDLE handle;
	char outname[NBUF];
	HEADER_INFO header_info;

	/* step 1: deal with command-line argument */
	if ((argc < 2) || (argc > 3))  {
		error(-1, USAGE);
	}
	if (strcmp(argv[1], "help") == 0) {
		printf("%s\n", USAGE);
		exit(0);
	}
	if ((fin = fopen(argv[1], "rb")) == NULL) {
		error(-1, "can't open input file");	
	}
	strcpy(outname, argv[2]);

	/* step 2: look in the input file, read a bunch of information */
	if (read_header_info(fin, &header_info) != 0) {
		error(-1, "invalid SBIG header information");
	}

	/* step 3: create the FITS file, write the appropriate header info */
	handle = fits_open(outname, "w", &(header_info.ncol), &(header_info.nrow));
	if (fill_fits_header(handle, &header_info) != 0) {
		error(-1, "cannot fill FITS header");
	}

#if 1
	/* step 4: copy the pixel data from SBIG to FITS */
	if (copy_pixel_data(fin, handle, &header_info) != 0) {
		error(-1, "error copying pixel data");
	}
#endif

	/* step 5: all done */
	fits_close(handle);

	exit(0);
}
		


	

	/***********************************************************************
	 * PROCEDURE: read_header_info
	 *
	 * DESCRIPTION: Read a bunch of information from the header of 
	 *              an SBIG image file.  Fill in the appropriate
	 *              fields of the given "header_info" structure.
	 *
	 *              We assume that a header line with the single word
	 *              "End" (spelled exactly that way) indicates the 
	 *              end of the header.
	 *
	 * RETURNS: 
	 *     0        if all goes well
	 *     1        if an error occurs
	 */

#define LINELEN  256

static int 
read_header_info
	(
	FILE *fin,              /* I: FILE pointer to SBIG file */
	HEADER_INFO *info       /* O: structure to hold header information */
	)
{
	int i, len;
	char keyword[LINELEN+1];
	char value[LINELEN+1];
	char line[LINELEN+1];

	rewind(fin);

	while (1) {
		if (fgets(line, LINELEN, fin) == NULL) {
			error(1, "read_header_info: unexpected end of file");
		}
		/* 
		 * if the first character of the line is carriage return \r, 
		 * remove it and shift all other characters backwards.
		 */
		if (line[0] == '\r') {
			len = strlen(line);
			for (i = 1; i < len; i++) {
				line[i - 1] = line[i];
			}
		}
		if (strncmp(line, "End", 3) == 0) {
			break;
		}
		if (sscanf(line, "%s = %s", keyword, value) == 2) {
#ifdef DEBUG
			printf("keyword ..%s.. value ..%s..\n", keyword, value);
#endif
		}

		/* 
		 * check to see if this keyword is important.  If so,
		 * save its value in the structure
		 */
		if (strncmp(keyword, "Width", 5) == 0) {
			info->nrow = atoi(value);
#ifdef DEBUG
			printf ("nrow is %d\n", info->nrow);
#endif
		}
		if (strncmp(keyword, "Height", 6) == 0) {
			info->ncol = atoi(value);
#ifdef DEBUG
			printf ("ncol is %d\n", info->ncol);
#endif
		}
		if (strncmp(keyword, "Date", 4) == 0) {
			sprintf(info->dateobs, "'%s'", value);
#ifdef DEBUG
			printf ("dateobs is ..%s..\n", info->dateobs);
#endif
		}
		if (strncmp(keyword, "Time", 4) == 0) {
			sprintf(info->timeobs, "'%s'", value);
#ifdef DEBUG
			printf ("timeobs is ..%s..\n", info->timeobs);
#endif
		}
		if (strncmp(keyword, "Filter", 6) == 0) {
			sprintf(info->filter, "'%s'", value);
#ifdef DEBUG
			printf ("filter is ..%s..\n", info->filter);
#endif
		}
		if (strcmp(keyword, "Exposure") == 0) {
			info->exptime = atof(value);
			/* convert from hundredths of a second to seconds */
			info->exptime /= 100.0;
#ifdef DEBUG
			printf ("exptime is %f\n", info->exptime);
#endif
		}
		if (strncmp(keyword, "Temperature", 10) == 0) {
			info->ccd_temp = atof(value);
#ifdef DEBUG
			printf ("ccd_temp is %f\n", info->ccd_temp);
#endif
		}
		if (strncmp(keyword, "E_gain", 6) == 0) {
			info->ccd_gain = atof(value);
#ifdef DEBUG
			printf ("ccd_gain is %f\n", info->ccd_gain);
#endif
		}
	}

	return(0);

}


/************************************************************************
 * PROCEDURE: fill_fits_header
 *
 * DESCRIPTION: We have already created a FITS file, and have a handle
 *              to it.  Use the handle to fill the FITS header with
 *              information from the given header_info structure.
 *              
 *              This is a little awkward because we must convert 
 *              numerical values into string format before calling
 *              the "fits_put_symbol" value.  
 *
 * RETURNS:
 *     0          if all goes well
 *     1          if an error occurs
 */


static int
fill_fits_header
	(
	FITS_HANDLE handle,     /* I: handle to output FITS file */
	HEADER_INFO *info       /* I: pointer to struct full of information */
	)
{
	char value[VAL_LEN];

	fits_put_symbol(handle, "DATE-OBS", info->dateobs);
	fits_put_symbol(handle, "TIME-OBS", info->timeobs);
	fits_put_symbol(handle, "FILTER", info->filter);

	sprintf(value, "%7.2f", info->exptime);
	fits_put_symbol(handle, "EXPTIME", value);

	sprintf(value, "%7.1f", info->ccd_temp);
	fits_put_symbol(handle, "CAM-TEMP", value);

	sprintf(value, "%7.2f", info->ccd_gain);
	fits_put_symbol(handle, "GAIN", value);


	return(0);
}



/************************************************************************
 * PROCEDURE: copy_pixel_data
 *
 * DESCRIPTION: Copy all the pixel values from the SBIG file to
 *              the new FITS file.  We translate from the SBIG
 *              convention of storing pixel values to the FITS
 *              convention here.  
 *
 * RETURNS:
 *    0            if all goes well
 *    1            if there's a problem
 */

static int
copy_pixel_data
	(
	FILE *fin,              /* I: FILE pointer to input SBIG file */
	FITS_HANDLE handle,     /* I: handle to output FITS file */
	HEADER_INFO *info       /* I: structure with information about image */
	)
{
	char pixel_bytes[2];
	char dummy_bytes[NMAX*2];
	int high, low;
	int nrow, ncol, row, col;
	int value, actual_value;
	int bytes_in_col;
	int pixel_value, code, last_pixel_value;
	int count_row, count_col, done, column_done;
	int16 fits_value;
	static int16 data[NMAX];


	nrow = info->nrow;
	ncol = info->ncol;

#ifdef DEBUG
	printf(" entering copy_pixel_data, nrow %5d ncol %5d \n",
				nrow, ncol);
#endif

	/* seek to the start of the data portion of the SBIG file */
	fseek(fin, (long) SBIG_DATA_START, SEEK_SET);




#if 0

	for (row = 0; row < nrow; row++) {

		for (col = 0; col < ncol; col++) {
			/* read in one pixel's data */
			if (fread(pixel_bytes, 1, 2, fin) != 2) {
				error(-1, "reached end of file while reading pixel data");
			}

			/* calculate the appropriate pixel value, in unsigned 16-bit */
			high = (unsigned char) pixel_bytes[1];
			low  = (unsigned char) pixel_bytes[0];
			value = -32768 + 256*high + low;
			if (value < MIN_DATA_VAL) {
				fits_value = MIN_DATA_VAL;
			} else if (value > MAX_DATA_VAL) {
				fits_value = MAX_DATA_VAL;
			} else {
				fits_value = value;
			}

			/* set the appropriate element of the 'data' array */
			data[col] = fits_value;
		}
		fits_put_data_fast(handle, data, ncol);
	
	}

#else

	/* 
	 * now, what I normally call "rows" are actually the columns
	 * in this compressed format.  So, I'll count as we read
	 * each _column_.  Each column contains data on a number
	 * of _rows_.
	 */

	count_col = 0;
	done = 0;
	while (done == 0) {

		/* 
		 * we are at the start of a line.  
		 * Read the first two bytes -- they tell us how many
		 * bytes long this compressed line is --
		 * which means how many "rows" in this column.
		 */
		/* read in one pixel's data */
		if (fread(pixel_bytes, 1, 2, fin) != 2) {
			error(-1, "reached end of file while reading pixel data");
		}
		/* calculate the value, in unsigned 16-bit */
		high = (unsigned char) pixel_bytes[1];
		low  = (unsigned char) pixel_bytes[0];
		bytes_in_col = 256*high + low;

		/* print it out */
#ifdef DEBUG
		printf(" line %5d has %6d bytes \n", count_col, bytes_in_col);
#endif

		/* 
		 * Now we must examine very carefully the data within
		 * this column.  There ought to be "nrow" data values
		 * encoded within these next bytes_in_col bytes.
		 */
		count_row = 0;
		
		/* 
		 * the first two bytes are always the actual data value
		 * of the first row in this column 
		 */
		if (fread(pixel_bytes, 1, 2, fin) != 2) {
			error(-1, "reached end of file while reading pixel data");
		}
		/* calculate the value, in unsigned 16-bit */
		high = (unsigned char) pixel_bytes[1];
		low  = (unsigned char) pixel_bytes[0];
		pixel_value = -32768 + 256*high + low;
#ifdef DEBUG
		printf("    pix in row %5d  is %6d \n", count_row, pixel_value);
#endif
		if (pixel_value < MIN_DATA_VAL) {
			fits_value = MIN_DATA_VAL;
		} else if (pixel_value > MAX_DATA_VAL) {
			fits_value = MAX_DATA_VAL;
		} else {
			fits_value = pixel_value;
		}
		/* set the appropriate element of the 'data' array */
		data[count_row] = fits_value;
		count_row++;

		last_pixel_value = pixel_value;

		/* 
		 * from this point on, there are two possibilities for
		 * the remaining data in this column.
		 *
		 *   a) if the next byte lies within range -127 to +127,
		 *        then the next pixel value is
		 *             new_value = last_value + delta
		 *        where "delta" is this byte
		 *
		 *   b) if the next byte is hex 0x80, then the next pixel
		 *        value is given by the following two bytes.
		 *
		 */
		column_done = 0;
		while (column_done == 0) {

#ifdef DEBUG
			printf("  col %6d row %6d  top of column_done loop \n", 
						count_col, count_row);
#endif

			/* read the next byte */
			if (fread(dummy_bytes, 1, 1, fin) != 1) {
				error(-1, "reached end of file while reading code pixel data");
			}
			
			code = (int) dummy_bytes[0];
#ifdef DEBUG
			printf("  code = %6d \n", code);
#endif


			if ((code >= -127) && (code <= 127)) {
				/* does it lie within the valid "delta" range? */

				pixel_value = last_pixel_value + code;
#ifdef DEBUG
				printf("  code = %6d, so next is %6d + %6d = %6d \n",
						code, last_pixel_value, code, pixel_value);
#endif

				/*do stuff here */
				if (pixel_value < MIN_DATA_VAL) {
					fits_value = MIN_DATA_VAL;
				} else if (pixel_value > MAX_DATA_VAL) {
					fits_value = MAX_DATA_VAL;
				} else {
					fits_value = pixel_value;
				}
				/* set the appropriate element of the 'data' array */
				data[count_row] = fits_value;


				last_pixel_value += code;
			}
			else if (code == -128) {
				/* or does it mean the pixel value lies in following two bytes? */
#ifdef DEBUG
				printf("  code = %6d, so next is as yet unknown \n",
						code);
#endif
				/* read in one pixel's data */
				if (fread(pixel_bytes, 1, 2, fin) != 2) {
					error(-1, "reached end of file while reading pixel data");
				}
				/* calculate the value, in unsigned 16-bit */
				high = (unsigned char) pixel_bytes[1];
				low  = (unsigned char) pixel_bytes[0];
				pixel_value = -32768 + 256*high + low;
#ifdef DEBUG
				printf("  next pixel_value is %6d \n", pixel_value);
#endif
				if (pixel_value < MIN_DATA_VAL) {
					fits_value = MIN_DATA_VAL;
				} else if (pixel_value > MAX_DATA_VAL) {
					fits_value = MAX_DATA_VAL;
				} else {
					fits_value = pixel_value;
				}
				/* set the appropriate element of the 'data' array */
				data[count_row] = fits_value;

				last_pixel_value = pixel_value;
	
			}
			else {
				/* huh?  Don't know what this means */
				fprintf(stderr, "copy_pixel_data: col %6d row %6d bad code %6d \n",
						count_col, count_row, code);
				error(-1, "bad code value, quitting ");
			}

			count_row++;
			if (count_row >= nrow) {
				column_done = 1;
			}
		}

#if 0
		fits_put_data_fast(handle, data, nrow);
#else
		fits_put_data(handle, count_col, 0, data, nrow);
#endif


		count_col++;
		if (count_col == ncol) {
			done = 1;
		}

	}


#endif


	return(0);
}
