#!/usr/bin/perl -w
#
# run some tests of a few of the XVista programs, checking the
#   results against known values.
#   
#   Print error messages as we go (if errors occur),
#   and exit with code
#
#        0           if all goes well
#        1           if error(s) occur
#        
# MWR 7/25/2011
#
# Added tests of the "star" program, with and without pmask.
#
# MWR 7/26/2011
#
# Fixed tests of the "star" program by changing FWHM of generated
#     stars from 3.0 to 4.0 pixels.  The FWHM = 3.0 was causing some
#     stars to be too "sharp" to qualify as real stars.
#
# MWR 6/23/2016
#

use strict;
use POSIX;

# set this to 1 to enable lots of debugging messages
my $debug = 1;

# we accept two numerical values as equal if they are the
#   same to within this amount
my $TOLERANCE = 1.0e-6;

# When we're running 'make', the "srcdir" is where the input files 
#   for the self-test live.
# But when we're running the test manually, the input files are
#   in the current directory
my $srcdir = $ENV{"srcdir"};
if (!defined($srcdir) || (length($srcdir) < 1)) {
  $srcdir = ".";
}
if ($debug > 0) {
  printf "srcdir is $srcdir\n";
}


my $temp_a = "$srcdir/selfa.dat";
my $temp_b = "$srcdir/selfb.dat";
my $temp_c = "$srcdir/selfc.dat";

# this will be the final exit code -- if 0, all is well.  
#   We increment it every time a test fails
my $final_code = 0;


my($teststr, $retval, @words);
my($expect_result);
my($mean);


######################################################
# First test: check that using the "pmask" feature
#    of the "add" program works properly.
#
#   create a small test image and set all pixels to 100
my $nrow = 100;
my $ncol = 100;
my $pixval = 100;

if ($debug > 0) {
  printf "  creating a small image with value %f \n", $pixval;
}
$teststr = "./pstar nrow=$nrow ncol=$ncol sky=100 outfile=$temp_a";
$retval = exec_cmd($teststr);
#   make sure that the file exists and that the mean value is 100
$teststr = "./mn $temp_a";
$retval = exec_cmd($teststr);
@words = split(/\s+/, $retval);
$mean = $words[0];
$mean =~ s/mean=//;
$final_code += check_equals($mean, $pixval);

#   now create a "mask" file which will be all zeroes
$teststr = "./pstar nrow=$nrow ncol=$ncol sky=0 outfile=$temp_b";
$retval = exec_cmd($teststr);
#   add a constant to the test image, using the mask.
#      Since the mask is all zeroes, we should add const to all pixels
if ($debug > 0) {
  printf "  testing that empty mask works with add \n";
}
$teststr = "./add $temp_a const=$pixval pmask=$temp_b outfile=$temp_c";
$retval = exec_cmd($teststr);
$teststr = "./mn $temp_c";
$retval = exec_cmd($teststr);
@words = split(/\s+/, $retval);
$mean = $words[0];
$mean =~ s/mean=//;
$expect_result = $pixval + $pixval;
$final_code += check_equals($mean, $expect_result);

#   now create a "mask" file which will is half zeros, half ones
$teststr = "./pstar nrow=$nrow ncol=$ncol sky=0 outfile=$temp_b";
$retval = exec_cmd($teststr);
$teststr = "./box 1 sr=0 sc=0 nr=$nrow/2 nc=$ncol ";
$retval = exec_cmd($teststr);
$teststr = "./add $temp_b const=1 box=1 ";
$retval = exec_cmd($teststr);
$teststr = "./mn $temp_b";
$retval = exec_cmd($teststr);
@words = split(/\s+/, $retval);
$mean = $words[0];
$mean =~ s/mean=//;
$expect_result = 0.5;
$final_code += check_equals($mean, $expect_result);

#   add a constant to the test image, using the mask.
#      Since the mask is now half 0 and half 1, we should
#      end up with an image which is 1.5 times original value
if ($debug > 0) {
  printf "  testing that non-empty mask works with add \n";
}
$teststr = "./add $temp_a const=$pixval pmask=$temp_b outfile=$temp_c";
$retval = exec_cmd($teststr);
$teststr = "./mn $temp_c";
$retval = exec_cmd($teststr);
@words = split(/\s+/, $retval);
$mean = $words[0];
$mean =~ s/mean=//;
$expect_result = 1.5*$pixval;
$final_code += check_equals($mean, $expect_result);



######################################################
# Next test: check that using the "pmask" feature
#    of the "sub" program works properly.
#
#   create a small test image and set all pixels to 100
if ($debug > 0) {
  printf "  creating a small image with value %f \n", $pixval;
}
$nrow = 100;
$ncol = 100;
$pixval = 100;
$teststr = "./pstar nrow=$nrow ncol=$ncol sky=100 outfile=$temp_a";
$retval = exec_cmd($teststr);
#   make sure that the file exists and that the mean value is 100
$teststr = "./mn $temp_a";
$retval = exec_cmd($teststr);
@words = split(/\s+/, $retval);
$mean = $words[0];
$mean =~ s/mean=//;
$final_code += check_equals($mean, $pixval);

#   now create a "mask" file which will be all zeroes
$teststr = "./pstar nrow=$nrow ncol=$ncol sky=0 outfile=$temp_b";
$retval = exec_cmd($teststr);
#   subtract a constant to the test image, using the mask.
#      Since the mask is all zeroes, we should sub const from all pixels
if ($debug > 0) {
  printf "  testing that empty mask works with sub \n";
}
$teststr = "./sub $temp_a const=$pixval pmask=$temp_b outfile=$temp_c";
$retval = exec_cmd($teststr);
$teststr = "./mn $temp_c";
$retval = exec_cmd($teststr);
@words = split(/\s+/, $retval);
$mean = $words[0];
$mean =~ s/mean=//;
$expect_result = $pixval - $pixval;
$final_code += check_equals($mean, $expect_result);

#   now create a "mask" file which will is half zeros, half ones
$teststr = "./pstar nrow=$nrow ncol=$ncol sky=0 outfile=$temp_b";
$retval = exec_cmd($teststr);
$teststr = "./box 1 sr=0 sc=0 nr=$nrow/2 nc=$ncol ";
$retval = exec_cmd($teststr);
$teststr = "./add $temp_b const=1 box=1 ";
$retval = exec_cmd($teststr);
$teststr = "./mn $temp_b";
$retval = exec_cmd($teststr);
@words = split(/\s+/, $retval);
$mean = $words[0];
$mean =~ s/mean=//;
$expect_result = 0.5;
$final_code += check_equals($mean, $expect_result);

#   sub a constant from the test image, using the mask.
#      Since the mask is now half 0 and half 1, we should
#      end up with an image which is 1.5 times original value
if ($debug > 0) {
  printf "  testing that non-empty mask works with sub \n";
}
$teststr = "./sub $temp_a const=$pixval pmask=$temp_b outfile=$temp_c";
$retval = exec_cmd($teststr);
$teststr = "./mn $temp_c";
$retval = exec_cmd($teststr);
@words = split(/\s+/, $retval);
$mean = $words[0];
$mean =~ s/mean=//;
$expect_result = 0.5*$pixval;
$final_code += check_equals($mean, $expect_result);




######################################################
# Next test: check that using the "pmask" feature
#    of the "mul" program works properly.
#
#   create a small test image and set all pixels to 100
if ($debug > 0) {
  printf "  creating a small image with value %f \n", $pixval;
}
$nrow = 100;
$ncol = 100;
$pixval = 100;
$teststr = "./pstar nrow=$nrow ncol=$ncol sky=100 outfile=$temp_a";
$retval = exec_cmd($teststr);
#   make sure that the file exists and that the mean value is 100
$teststr = "./mn $temp_a";
$retval = exec_cmd($teststr);
@words = split(/\s+/, $retval);
$mean = $words[0];
$mean =~ s/mean=//;
$final_code += check_equals($mean, $pixval);

#   now create a "mask" file which will be all zeroes
$teststr = "./pstar nrow=$nrow ncol=$ncol sky=0 outfile=$temp_b";
$retval = exec_cmd($teststr);
#   multiply the test image by a constant, using the mask.
#      Since the mask is all zeroes, we should mult all pixels by const
if ($debug > 0) {
  printf "  testing that empty mask works with mul \n";
}
$teststr = "./mul $temp_a const=2 pmask=$temp_b outfile=$temp_c";
$retval = exec_cmd($teststr);
$teststr = "./mn $temp_c";
$retval = exec_cmd($teststr);
@words = split(/\s+/, $retval);
$mean = $words[0];
$mean =~ s/mean=//;
$expect_result = $pixval*2.0;
$final_code += check_equals($mean, $expect_result);

#   now create a "mask" file which will is half zeros, half ones
$teststr = "./pstar nrow=$nrow ncol=$ncol sky=0 outfile=$temp_b";
$retval = exec_cmd($teststr);
$teststr = "./box 1 sr=0 sc=0 nr=$nrow/2 nc=$ncol ";
$retval = exec_cmd($teststr);
$teststr = "./add $temp_b const=1 box=1 ";
$retval = exec_cmd($teststr);
$teststr = "./mn $temp_b";
$retval = exec_cmd($teststr);
@words = split(/\s+/, $retval);
$mean = $words[0];
$mean =~ s/mean=//;
$expect_result = 0.5;
$final_code += check_equals($mean, $expect_result);

#   mul the test image by a constant, using the mask.
#      Since the mask is now half 0 and half 1, we should
#      end up with an image which is 1.5 times original value
if ($debug > 0) {
  printf "  testing that non-empty mask works with mul \n";
}
$teststr = "./mul $temp_a const=2 pmask=$temp_b outfile=$temp_c";
$retval = exec_cmd($teststr);
$teststr = "./mn $temp_c";
$retval = exec_cmd($teststr);
@words = split(/\s+/, $retval);
$mean = $words[0];
$mean =~ s/mean=//;
$expect_result = 1.5*$pixval;
$final_code += check_equals($mean, $expect_result);




######################################################
# Next test: check that using the "pmask" feature
#    of the "div" program works properly.
#
#   create a small test image and set all pixels to 100
if ($debug > 0) {
  printf "  creating a small image with value %f \n", $pixval;
}
$nrow = 100;
$ncol = 100;
$pixval = 100;
$teststr = "./pstar nrow=$nrow ncol=$ncol sky=100 outfile=$temp_a";
$retval = exec_cmd($teststr);
#   make sure that the file exists and that the mean value is 100
$teststr = "./mn $temp_a";
$retval = exec_cmd($teststr);
@words = split(/\s+/, $retval);
$mean = $words[0];
$mean =~ s/mean=//;
$final_code += check_equals($mean, $pixval);

#   now create a "mask" file which will be all zeroes
$teststr = "./pstar nrow=$nrow ncol=$ncol sky=0 outfile=$temp_b";
$retval = exec_cmd($teststr);
#   divide the test image by a constant, using the mask.
#      Since the mask is all zeroes, we should mult all pixels by const
if ($debug > 0) {
  printf "  testing that empty mask works with mul \n";
}
$teststr = "./div $temp_a const=2 pmask=$temp_b outfile=$temp_c";
$retval = exec_cmd($teststr);
$teststr = "./mn $temp_c";
$retval = exec_cmd($teststr);
@words = split(/\s+/, $retval);
$mean = $words[0];
$mean =~ s/mean=//;
$expect_result = $pixval/2.0;
$final_code += check_equals($mean, $expect_result);

#   now create a "mask" file which will is half zeros, half ones
$teststr = "./pstar nrow=$nrow ncol=$ncol sky=0 outfile=$temp_b";
$retval = exec_cmd($teststr);
$teststr = "./box 1 sr=0 sc=0 nr=$nrow/2 nc=$ncol ";
$retval = exec_cmd($teststr);
$teststr = "./add $temp_b const=1 box=1 ";
$retval = exec_cmd($teststr);
$teststr = "./mn $temp_b";
$retval = exec_cmd($teststr);
@words = split(/\s+/, $retval);
$mean = $words[0];
$mean =~ s/mean=//;
$expect_result = 0.5;
$final_code += check_equals($mean, $expect_result);

#   divide the test image by a constant, using the mask.
#      Since the mask is now half 0 and half 1, we should
#      end up with an image which is 0.75 times orig value
if ($debug > 0) {
  printf "  testing that non-empty mask works with mul \n";
}
$teststr = "./div $temp_a const=2 pmask=$temp_b outfile=$temp_c";
$retval = exec_cmd($teststr);
$teststr = "./mn $temp_c";
$retval = exec_cmd($teststr);
@words = split(/\s+/, $retval);
$mean = $words[0];
$mean =~ s/mean=//;
$expect_result = 0.75*$pixval;
$final_code += check_equals($mean, $expect_result);



######################################################
# Test the "stars" program -- 
#    verify that it can find a single star in a sparse image
#    verify that the "pmask=" option prevents it from
#        finding that star
#
#   create a small test image which has a known sky value,
#      random noise, and a single bright star near center
my $sky = 100;
my $skysig = sqrt($sky);
my $star_cr = $nrow/2;
my $star_cc = $ncol/2;
my $star_peak = 1000;
my $star_fwhm = 4.0;
my $ret_code;

if ($debug > 0) {
  printf "  testing 'stars' without pmask ... \n";
}
$teststr = "./pstar nrow=$nrow ncol=$ncol sky=$sky outfile=$temp_a ";
$teststr .= " star=$star_peak,$star_cr,$star_cc,$star_fwhm ";
$retval = exec_cmd($teststr);
#   make sure that "stars" can find the single star
$teststr = "./stars $temp_a sky=$sky skysig=$skysig minsig=10 quiet";
$retval = exec_cmd($teststr);
#       here we check that there is a single line
@words = split(/\n/, $retval);
if ($#words != 0) {
  if ($debug > 0) {
    printf "   stars test fails: nwords %d != 0 \n", $#words;
  }
  $final_code++;
}
#       here we check that the single star is in the right place
@words = split(/\s+/, $retval);
my $row = $words[2];
$ret_code = check_equals($row, $star_cr, 0.5);
if ($ret_code != 0) {
  printf "   stars test fails: row %.2f star_cr %.2f \n",
          $row, $star_cr;
}
$final_code += $ret_code;
my $col = $words[3];
$ret_code = check_equals($col, $star_cc, 0.5);
if ($ret_code != 0) {
  printf "   stars test fails: col %.2f star_cc %.2f \n",
          $col, $star_cc;
}
$final_code += $ret_code;

if ($debug > 0) {
  printf "  testing 'stars' with pmask ... \n";
}
#   we make a pmask which is zero except for a central
#      block of 1s at the position of the star
my $box_num = 9;
$teststr = "cp $temp_a $temp_b";
$retval = exec_cmd($teststr);
$teststr = "./box $box_num cr=$star_cr cc=$star_cc nr=10 nc=10 ";
$retval = exec_cmd($teststr);
$teststr = "./mul $temp_b const=0 ";
$retval = exec_cmd($teststr);
$teststr = "./add $temp_b const=1 box=$box_num ";
$retval = exec_cmd($teststr);
#   now, with this mask, we should NOT find the star
$teststr = "./stars $temp_a sky=$sky skysig=$skysig minsig=10 quiet";
$teststr .= " pmask=$temp_b ";
$retval = exec_cmd($teststr);
#       here we check that there should be no lines
if (length($retval) != 0) {
  printf "   stars test fails: found star when using mask?! \n";
  printf "   %s ", $retval;
  $final_code++;
}



# if not in debug mode, remove all the temporary files
if ($debug == 0) {
  unlink($temp_a);
  unlink($temp_b);
  unlink($temp_c);
}

if ($debug > 0) {
  printf " exiting with code $final_code \n";
}
exit $final_code;




##############################################################################
# PROCEDURE: check_equals
#
# DESCRIPTION: Check to see if the two numerical arguments
#              are the same, to within the given tolerance.
#              If no tolerance is given, use $TOLERANCE.  
#              
# RETURNS:     0        if they are the same
#              1        if not the same
#
sub check_equals {
  my($a, $b, $tol_arg);

  $a = $_[0];
  $b = $_[1];
  $tol_arg = $_[2];

  if (!defined($tol_arg) || ($tol_arg == 0)) {
    $tol_arg = $TOLERANCE;
  }

  if (fabs($a - $b) < $tol_arg) {
    return(0);
  }
  else {
    return(1);
  }
}


##############################################################################
# PROCEDURE: exec_cmd
#
# DESCRIPTION: Execute the given shell command line.  If the $debug flag
#              is set, we print to stdout the command line, and
#              also print out the result string.
#
# RETURNS:
#              the result of the command
#
#
sub exec_cmd {

  my($cmd, $ret);

  $cmd = $_[0];

  if ($debug > 0) {
    printf "cmd is ..$cmd.. \n";
  }
  $ret = `$cmd`;
  if ($debug > 0) {
    printf "ret is ..$ret.. \n";
  }

  return($ret);
}


