
	/*	
	 * Reads a raw Tomoe-e FITS image, which has a 3-D format
	 * with NAXIS = 3.  What we want to do is split this
	 * single file, which contains many 2-D images,
	 * into a set of many files, each of which contains ONE 2-D image.
	 * 
	 *    Usage:   split3d   infile  outfile=base
	 *
	 * where
	 *                 infile          is original raw KWFC FITS image
	 *  
	 *                 base            is base of name for new output files
	 *                                   will have names like base_0001.fts
	 *                                                        base_0002.fts
	 *                                                        ...
	 *                                                        base_0nnn.fts
	 *        
	 *
	 * MWR 10/18/2016
	 *
	 * Modified to handle a newer type of Tomo-e FITS file, as well as
	 * the original version.
	 *              - original has BITPIX=16 
	 *              - new version has BITPIX=-32
	 *
	 * MWR 6/3/2019
	 */



#include "pcvista.h"
#include "fits.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

#undef DEBUG
#undef DEBUG2

#define MAX_HEADER_LINES 1000
#define CARD_LENGTH      80
#define CARDS_PER_BUNCH  36



char *progname = "split3d";

#define USAGE  "usage: split3d  file outfile=  [help=] "


/* 
 * global variables, used by both
 *     check_input_header()
 *     read_and_convert()
 */
char header_array[MAX_HEADER_LINES + 1][CARD_LENGTH + 1];
char bits_per_pix_16_string[BIG_BUF];
char bits_per_pix_32_string[BIG_BUF];


static int check_input_header(
	char *filename,        /* I: name of FITS file  */
	int *naxis1,           /* O: value of NAXIS1 from header */
	int *naxis2,           /* O: value of NAXIS2 from header */
	int *naxis3,           /* O: value of NAXIS3 from header */
	int *bitpix,           /* O: value of BITPIX from header */
	int *num_cards         /* O: number of non-blank FITS header cards */
	);

static int
read_and_convert
	(
	char *infile,          /* I: name of input FITS file  */
	char *outfile,         /* I: name of output FITS file (we create it) */
	int naxis1,            /* I: value of NAXIS1 from header */
	int naxis2,            /* I: value of NAXIS2 from header */
	int naxis3,            /* I: value of NAXIS3 from header */
	int bitpix,            /* I: value of BITPIX from header */
	int num_cards          /* I: number of non-empty FITS header cards */
	);

static int
decode_pixel_value
	(
	unsigned char low_byte,         /* I: low byte of pixel data */
	unsigned char high_byte,        /* I: high byte of pixel data */
	double bzero                    /* I: BZERO from FITS header */
	);

static int
encode_pixel_value
	(
	int pixel_value,                /* I: encode this value */
	unsigned char *low_byte,        /* O: low byte of encoded pixel data */
	unsigned char *high_byte        /* O: high byte of encoded pixel data */
	);

static int
pad_headerline
	(
	char *headerline              /* I/O: characters to go in FITS header */
	);

static int
calc_data_block_size
	(
	int naxis1,       /* I: number of columns of data in each image */
	int naxis2,       /* I: number of rows of data in each image */
	int bitpix        /* I: number of bits per pixel (16 or -32) */
	);




int
main(argc,argv)
int argc;
char *argv[];
{
	char infile[BIG_BUF], outfile[BIG_BUF];
	char *gotit;
	int i;
	int naxis1, naxis2, naxis3, bitpix, num_cards;
	

	if (argc == 1) {
		fprintf(stderr, "%s\n", USAGE);
		exit(1);
	}
	if (strcmp(argv[1], "help") == 0) {
		printf("%s\n", USAGE);
		exit(0);
	}

	/* prevent string overruns */
	infile[BIG_BUF - 1] = '\0';
	outfile[BIG_BUF - 1] = '\0';
	
	/* initialize stuff */
	strcpy(infile, "");
	strcpy(outfile, "");
	naxis1 = -1;
	naxis2 = -1;
	naxis3 = -1;
	
	/* the first argument will be the input file name */
	strncpy(infile, argv[1], BIG_BUF-1);

	for (i = 2; i < argc; i++) {
		if (keyword("help", argv[i])) {
			printf("%s\n", USAGE);
			exit(0);
		}
		if ((gotit = find("outfile", argv[i])) != NULL) {
			strncpy(outfile, gotit, BIG_BUF-1);
			continue;
		}
	}
#ifdef DEBUG
	printf(" infile %s  outfile %s \n", infile, outfile);
#endif
	
	/* check arguments for validity before continuing */
	if ((strlen(infile) == 0) || (strlen(outfile) == 0)) {
		fprintf(stderr, "invalid input and/or output file names \n");
		exit(1);
	}
	/* 
	 * watch out for too large a length of outfile -- we are
	 *    going to extend the output file name with something like
	 *     
	 *           _0435.fits
	 *
	 *    and we don't want to overflow a buffer.
	 */
	{ 
		int maxlen = BIG_BUF - 20;

		if (strlen(outfile) >= maxlen) {
			fprintf(stderr, "%s: outfile name is longer than %d chars \n",
						progname, maxlen);
			exit(1);
		}
	}


	/* read the input file header */
	if (check_input_header(infile, &naxis1, &naxis2, &naxis3, &bitpix,
				&num_cards) != 0) {
		/* an error occurred */
		fprintf(stderr, "%s: check_input_header fails \n", progname);
		exit(1);
	}
	else {
#ifdef DEBUG
		printf(" result of check_input_header: \n");
		printf(" naxis1 %5d  naxis2 %5d naxis3 %5d bitpix %5d \n", 
				naxis1, naxis2, naxis3, bitpix);
		printf(" num_cards %5d \n", num_cards);
#endif
	}


	/* read the input file data, convert and write into output file */
	if (read_and_convert(infile, outfile, naxis1, naxis2, naxis3, bitpix,
			num_cards) != 0) {
		fprintf(stderr, "%s: read_and_convert fails \n", progname);
		exit(1);
	}
	else {
#ifdef DEBUG
		printf("read_and_convert succeeds \n");
#endif
	}


	exit(0);
}


	/*************************************************************
	 * PROCEDURE: check_input_header
	 *
	 * DESCRIPTION: read the FITS header of the input file.
	 *              Verify that
	 *
	 *                   SIMPLE = T
	 *                   BITPIX = 16  or -32
	 *                   NAXIS  = 3
	 * 
	 *              Read the values of NAXIS1 and NAXIS2 and NAXIS3
	 *
	 *              Place results into output arguments.
	 *
	 * RETURNS:     0       if all goes well
	 *              1       if an error occurs
	 */


static int
check_input_header
	(
	char *filename,        /* I: name of FITS file  */
	int *naxis1,           /* O: value of NAXIS1 from header */
	int *naxis2,           /* O: value of NAXIS2 from header */
	int *naxis3,           /* O: value of NAXIS3 from header */
	int *bitpix,           /* O: value of BITPIX from header */
	int *num_cards         /* O: number of FITS header cards */
	                       /*     up to the "END" card */
	)
{
	int i;
	int done;
	int start_card, end_card, card_index;
	int naxis1_val, naxis2_val, naxis3_val, bitpix_val;
	int simple_flag, naxis_flag, naxis1_flag, naxis2_flag, naxis3_flag;
	int bitpix_flag;
	FILE *fp;
	


	/* 
	 * set all flags to zero -- signalling no information yet.
	 * As we read bit of info from the header, we check for valid value.
	 * If value is OK, we set flag to 1.
	 */
	simple_flag = 0;
	bitpix_flag = 0;
	naxis_flag = 0;
	naxis1_flag = 0;
	naxis2_flag = 0;
	naxis3_flag = 0;

	/* open file for reading */
	if ((fp = fopen(filename, "r")) == NULL) {
		fprintf(stderr, "check_input_header: can't open file %s \n",
				filename);
		return(1);
	}
		

	/* read in all header lines until we find an "END   " line */
	done = 0;
	start_card = 0;
	while (!done) {

#ifdef DEBUG
		printf("starting to read card %d \n", start_card);
#endif
		/* read in a "bunch" of header lines */
		for (i = 0; i < CARDS_PER_BUNCH; i++) {
			card_index = i + start_card;
			if (fread(header_array[card_index], CARD_LENGTH, 1, fp) != 1) {
				fprintf(stderr, "check_input_header: failed to read card %d \n", 
						card_index);
				return(1);
			}
			/* add a string terminator to end of card, for convenience */
			header_array[card_index][CARD_LENGTH] = '\0';
#ifdef DEBUG
			printf(" card %d is ..%s..\n", card_index, header_array[card_index]);
#endif

			/* check to see if we found the END card */
			if (strncmp(header_array[card_index], "END     ", 8) == 0) {
				end_card = card_index;
				done = 1;
			}

		}
		start_card += CARDS_PER_BUNCH;
	}
	fclose(fp);



	/* 
	 * If we get here, then we did succeed in reading the FITS header.
	 *
	 * Now, verify that this is a very simple, BITPIX=16 or BITPIX=-32 FITS image,
	 * and get the number of rows and cols and "snaps".
	 */
	for (i = 0; i <= end_card; i++) {
		char header_line_copy[CARD_LENGTH + 1];
		char *delim, *tok;
		delim = " ";

		/* make a copy of each line so strtok doesn't modify header lines */
		strncpy(header_line_copy, header_array[i], CARD_LENGTH);
		header_line_copy[CARD_LENGTH] = '\0';

		/* verify that SIMPLE = T */
		if (strncmp(header_line_copy, "SIMPLE  =", 9) == 0) {
			tok = strtok(header_line_copy + 9, delim);
#ifdef DEBUG
			printf("tok for SIMPLE is ..%s.. \n", tok);
#endif
			if (strcmp(tok, "T") != 0) {
				fprintf(stderr, "SIMPLE is not T \n");
				return(1);
			}
			simple_flag = 1;
			continue;
		}

		/* verify that BITPIX = 16 or -32 */
		if (strncmp(header_line_copy, "BITPIX  =", 9) == 0) {

			tok = strtok(header_line_copy + 9, delim);
#ifdef DEBUG
			printf("tok for BITPIX is ..%s.. \n", tok);
#endif
			bitpix_val = (int) strtol(tok, NULL, 10);
#ifdef DEBUG
			printf("we read bitpix = %d \n", bitpix_val);
#endif
			if ((bitpix_val != 16) && (bitpix_val != -32)) {
				fprintf(stderr, "BITPIX is not 16 or -32 \n");
				return(1);
			}
			bitpix_flag = 1;
			continue;
		}


		/* verify that NAXIS = 2 */
		if (strncmp(header_line_copy, "NAXIS   =", 9) == 0) {
			tok = strtok(header_line_copy + 9, delim);
#ifdef DEBUG
			printf("tok for NAXIS is ..%s.. \n", tok);
#endif
			if (strcmp(tok, "3") != 0) {
				fprintf(stderr, "NAXIS is not 3 \n");
				return(1);
			}
			naxis_flag = 1;
			continue;
		}

		/* get the value for  NAXIS1 */
		if (strncmp(header_line_copy, "NAXIS1  =", 9) == 0) {
			tok = strtok(header_line_copy + 9, delim);
#ifdef DEBUG
			printf("tok for NAXIS1 is ..%s.. \n", tok);
#endif
			naxis1_val = atoi(tok);
#ifdef DEBUG
			printf("naxis1_val is %d \n", naxis1_val);
#endif
			if ((naxis1_val <= 0) || (naxis1_val >= NMAX)) {
				fprintf(stderr, "invalid value for NAXIS1 %s \n", tok);
				return(1);
			}
			naxis1_flag = 1;
			continue;
		}

		/* get the value for  NAXIS2 */
		if (strncmp(header_line_copy, "NAXIS2  =", 9) == 0) {
			tok = strtok(header_line_copy + 9, delim);
#ifdef DEBUG
			printf("tok for NAXIS2 is ..%s.. \n", tok);
#endif
			naxis2_val = atoi(tok);
#ifdef DEBUG
			printf("naxis2_val is %d \n", naxis2_val);
#endif
			if ((naxis2_val <= 0) || (naxis2_val >= NMAX)) {
				fprintf(stderr, "invalid value for NAXIS2 %s \n", tok);
				return(1);
			}
			naxis2_flag = 1;
			continue;
		}

		/* get the value for  NAXIS3 */
		if (strncmp(header_line_copy, "NAXIS3  =", 9) == 0) {
			tok = strtok(header_line_copy + 9, delim);
#ifdef DEBUG
			printf("tok for NAXIS3 is ..%s.. \n", tok);
#endif
			naxis3_val = atoi(tok);
#ifdef DEBUG
			printf("naxis3_val is %d \n", naxis3_val);
#endif
			if ((naxis3_val <= 0) || (naxis3_val >= NMAX)) {
				fprintf(stderr, "invalid value for NAXIS3 %s \n", tok);
				return(1);
			}
			naxis3_flag = 1;
			continue;
		}


	}

	/* make sure that
	 *
	 *   a) we read all required information from header
	 *   b) all information we read was valid
	 */
	if ((simple_flag == 0) || (naxis_flag == 0) || 
	    (naxis1_flag == 0) || (naxis2_flag == 0) ||
	    (naxis3_flag == 0) || (bitpix_flag == 0)) {
		fprintf(stderr, "check_input_header: invalid information in header \n");
		return(1);
	}


	/* set the output args to the values we read from header */
	*naxis1 = naxis1_val;
	*naxis2 = naxis2_val;
	*naxis3 = naxis3_val;
	*bitpix = bitpix_val;
	*num_cards = end_card;

	return(0);
}



	/*************************************************************
	 * PROCEDURE: read_and_convert
	 *
	 * DESCRIPTION: We know the parameters of the input FITS image.
	 *              So, we can now 
	 *               
	 *              LOOP: i = 0 to NAXIS3
	 *                  read data section i
	 *                  write header + data section i to an output file
	 *
	 *              If BITPIX=16, then we can simply read bytes from the input
	 *              file and copy them directly to the output file; we don't
	 *              need to perform any translation.  But if BITPIX=-32,
	 *              we need to 
	 *                    - read in a bunch of 32-bit floating point numbers
	 *                    - check the value of each one, clipping to the
	 *                         maximum values of 16-bit integers if necessary,
	 *                         and setting elements of an output array
	 *                    - write the output array to the output file
	 *        
	 *              Also, of course, switch the value of BITPIX in the
	 *              FITS header from -32 to 16, if necessary.
	 *
	 *
	 * RETURNS:     0       if all goes well
	 *              1       if an error occurs
	 */


static int
read_and_convert
	(
	char *infile,          /* I: name of input FITS file  */
	char *outbase,         /* I: basename of output FITS file (we create it) */
	int naxis1,            /* I: value of NAXIS1 from header */
	int naxis2,            /* I: value of NAXIS2 from header */
	int naxis3,            /* I: value of NAXIS3 from header */
	int bitpix,            /* I: value of BITPIX from header */
	int num_cards          /* I: number of non-empty FITS header cards */
	)
{
	int i, index;
	int num_cards_written;
	char outfile[BIG_BUF];
	char headerline[CARD_LENGTH + 1];
	FILE *infp, *outfp;

	/* loop over all NAXIS3 values */
	for (index = 0; index < naxis3; index++) {


		/* open the input file */
		if ((infp = fopen(infile, "r")) == NULL) {
			fprintf(stderr, "read_and_convert: can't open input file %s \n",
					infile);
			return(1);
		}	


		/* create and open the output file */
		sprintf(outfile, "%s_%04d.fits", outbase, index);
		if ((outfp = fopen(outfile, "w")) == NULL) {
			fprintf(stderr, "read_and_convert: can't open output file %s \n",
					outfile);
			return(1);
		}	
	
		/* copy the FITS header, up to but not including END */
		for (i = 0; i < num_cards; i++) {
	
	#ifdef DEBUG
			printf("  next is card %5d \n", i);
			strncpy(headerline, header_array[i], CARD_LENGTH);
			headerline[CARD_LENGTH] = '\0';
			printf("%s\n", headerline);
	#endif
	
			/* If this is NAXIS3 card, we skip it */
			if (strncmp(header_array[i], "NAXIS3  =", 9) == 0) {
	#ifdef DEBUG
				printf(" skipping NAXIS3 line \n");
	#endif
				continue;
			}
	
			/* If this is NAXIS card, we change its value from 3 to 2 */
			if (strncmp(header_array[i], "NAXIS   =", 9) == 0) {
				sprintf(headerline, "NAXIS   = %20d ", 2);
				if (pad_headerline(headerline) != 0) {
					fprintf(stderr, "read_and_convert: failed to pad NAXIS \n");
					return(1);
				}
	#ifdef DEBUG
				printf(" printing the following modified NAXIS line\n");
				headerline[CARD_LENGTH] = '\0';
				printf("%s\n", headerline);
	#endif
				if (fwrite(headerline, CARD_LENGTH, 1, outfp) != 1) {
					fprintf(stderr, "read_and_convert: failed to write NAXIS \n");
					return(1);
				}
				continue;
			}

			/* If this is BITPIX card, we force its value to 16 */
			if (strncmp(header_array[i], "BITPIX  =", 9) == 0) {
				sprintf(headerline, "BITPIX  = %20d ", 16);
				if (pad_headerline(headerline) != 0) {
					fprintf(stderr, "read_and_convert: failed to pad BITPIX \n");
					return(1);
				}
	#ifdef DEBUG
				printf(" printing the following modified BITPIX line\n");
				headerline[CARD_LENGTH] = '\0';
				printf("%s\n", headerline);
	#endif
				if (fwrite(headerline, CARD_LENGTH, 1, outfp) != 1) {
					fprintf(stderr, "read_and_convert: failed to write BITPIX \n");
					return(1);
				}
				continue;
			}
	
	
			/* otherwise, just copy the card to output */
			if (fwrite(header_array[i], CARD_LENGTH, 1, outfp) != 1) {
	#ifdef DEBUG
				printf("  writing the line unmodified to header \n");
	#endif
				fprintf(stderr, "read_and_convert: failed to write card %d \n", 
						i);
				return(1);
			}
		}
	
		/* add a COMMENT line describing what we've done */
		sprintf(headerline, "COMMENT   split3d index %d ", index);
		if (pad_headerline(headerline) != 0) {
			fprintf(stderr, "read_and_convert: failed to pad COMMENT \n");
			return(1);
		}
		if (fwrite(headerline, CARD_LENGTH, 1, outfp) != 1) {
			fprintf(stderr, "read_and_convert: failed to write COMMENT \n");
			return(1);
		}
	
		/* add an END line */
		sprintf(headerline, "END");
		if (pad_headerline(headerline) != 0) {
			fprintf(stderr, "read_and_convert: failed to pad END \n");
			return(1);
		}
		if (fwrite(headerline, CARD_LENGTH, 1, outfp) != 1) {
			fprintf(stderr, "read_and_convert: failed to write END \n");
			return(1);
		}
	
		/* 
		 * now we write additional blank cards to bring the total number of
		 * cards up to a multiple of CARDS_PER_BUNCH
		 */
		num_cards_written = num_cards + 1;   /* may need to fix this */
	
		if (num_cards_written % CARDS_PER_BUNCH != 0) {
			int j;
			char blank_line[CARD_LENGTH];
	
			for (j = 0; j < CARD_LENGTH; j++) {
				blank_line[j] = ' ';
			}
	
			i = num_cards;
			while ((i + 1) % CARDS_PER_BUNCH != 0) {
	#ifdef DEBUG
				printf(" writing blank line, card %d \n", i);
	#endif
				if (fwrite(blank_line, CARD_LENGTH, 1, outfp) != 1) {
					fprintf(stderr, "read_and_convert: failed to write blank %d \n", 
							i);
					return(1);
				}
				i++;
				num_cards_written++;
			}
		}
	#ifdef DEBUG
		printf("num_cards_written is %d \n", num_cards_written);
	#endif
				

		/* sanity check: should be at boundary of 2880 bytes */
		{
			long pos;

			pos = ftell(outfp);
			if (pos % (CARD_LENGTH*CARDS_PER_BUNCH) != 0) {
				fflush(stdout);
				fprintf(stderr, "after writing header, at pos %ld \n", pos);
				return(1);
			}
		}
				
		/* copy the FITS data from the index'th image */
		{
			int row, col;
			char data_array[4*NMAX];
			char integer_data_array[2*NMAX];
			long offset, data_block_size;
			size_t bytes_per_pixel, num_rows, num_cols;
		
			/* seek to start of data in input file */
			offset = num_cards_written*CARD_LENGTH +
						calc_data_block_size(naxis1, naxis2, bitpix)*index;



	#ifdef DEBUG
			printf("seeking to position %ld in %s \n", offset, infile);
	#endif
			if (fseek(infp, offset, SEEK_SET) != 0) {
				fprintf(stderr, "read_and_convert: fseek to %ld in %s fails \n",
							offset, infile);
				return(1);
			}
		

			num_rows = naxis2;
			num_cols = naxis1;
	#ifdef DEBUG
			printf(" now_rows is %d  num_cols is %d \n", (int) num_rows, (int) num_cols);
	#endif

			/****************************************************************
 			 * If bitpix=16, we simply read a chunk of bytes and write the
 			 * same chunk to output file.  No need to examine data values.
			 */
			if (bitpix == 16) {
				bytes_per_pixel = 2;
				for (row = 0; row < num_rows; row++) {
	
					if (fread(data_array, bytes_per_pixel, num_cols, infp) != num_cols) {
						fprintf(stderr, "read_and_convert: fread fails row %d \n", row);
						return(1);
					}
					if (fwrite(data_array, bytes_per_pixel, num_cols, outfp) != num_cols) {
						fprintf(stderr, "read_and_convert: fwrite fails row %d \n", row);
						return(1);
					}
				}
			}

			/****************************************************************
 			 * If bitpix=-32, we must not only read a chunk of bytes for each
 			 *    row of pixels, we must then read the pixel values,
 			 *    and convert from floating point to integer, clipping
 			 *    values if necessary.  Then we can write the smaller
 			 *    array out 16-bit pixels to the output file.
			 */
			if (bitpix == -32) {
				char *p, *outp;
				int col, int_bytes_per_pixel;
				float *fp, float_data_array[NMAX];
			
				bytes_per_pixel = 4;
				int_bytes_per_pixel = 2;

				
				for (row = 0; row < num_rows; row++) {
	
					if (fread((char *) float_data_array, bytes_per_pixel, num_cols, infp) != num_cols) {
						fprintf(stderr, "read_and_convert: fread fails row %d \n", row);
						return(1);
					}

					/* 
 	 				 * here's where we translate the data from floating point
 	 				 * to 16-bit integer
 	 				 */
					p = (char *) &(float_data_array[0]);
					outp = (char *) &(integer_data_array[0]);
					for (col = 0; col < num_cols; col++) {

						/* 
 						 * we need to reverse the order of the bytes in each
 						 * set of 4 consecutive bytes 
 						 */
						char byte_array[4], *b;
						int16 new_value;
			
						b = byte_array;
						*b = *(p+3); *(b+1) = *(p+2); *(b+2) = *(p+1); *(b+3) = *p;

						/* make sure the result is in value range */
						fp = (float *) &byte_array[0];
						if ((*fp) < MIN_DATA_VAL) {
							new_value = MIN_DATA_VAL;
						}
						else if ((*fp) > MAX_DATA_VAL) {
							new_value = MAX_DATA_VAL;
						}
						else {
							/* 
							 * this rounds properly for positive values;
							 * for negative values, it yields a positive bias.
							 * But we don't care about negative values (they 
							 * appear only in bias regions), and we need the speed.
							 */
							new_value = (int16) ((*fp) + 0.5);
						}
#ifdef DEBUG
						printf(" col %5d  fval %f new_value %d \n", col, *fp, new_value);
#endif
						/*
                   * now swap the bytes of the new integer value 
                   * and place those swapped bytes into output array 
                   */
						{
							char *pp, *qq;
							int16 swapped_value;

							pp = (char *) &new_value;
							qq = (char *) &swapped_value;
							
							*qq = *(pp + 1);
							*(qq + 1) = *pp;

							*outp++ = *qq++;
							*outp++ = *qq++;
						}

						p += bytes_per_pixel;
					}


					/* 
 					 * now we can write out the converted pixel values
 					 */
					if (fwrite(integer_data_array, int_bytes_per_pixel, num_cols, outfp) != num_cols) {
						fprintf(stderr, "read_and_convert: fwrite fails row %d \n", row);
						return(1);
					}

				}
			}

		}
	
		/* 
		 * we must now pad the output file with NUL characters until we
		 * reach a multiple of CARD_LENGTH * CARDS_PER_BUNCH bytes total 
		 */
		{
			char pad_array[CARD_LENGTH*CARDS_PER_BUNCH + 1];
			int i;
			long total_bytes, pad_bytes, bunch_bytes;
	
			/* file must contain a multiple of this many bytes */
			bunch_bytes = CARD_LENGTH * CARDS_PER_BUNCH;
			total_bytes = ftell(outfp);
			pad_bytes = (((total_bytes/bunch_bytes) + 1)*bunch_bytes) - total_bytes;
			if (pad_bytes == bunch_bytes) {
				/* we ended exactly on a boundary by chance.  No need to pad */
#ifdef DEBUG
				printf("  no need to pad, output file at %ld bytes \n", total_bytes);
#endif
			}
			else {
#ifdef DEBUG
				printf("  after data, output file has %ld bytes \n", total_bytes);
				printf("  so add %ld bytes to pad it \n", pad_bytes);
#endif
				for (i = 0; i < pad_bytes; i++) {
					pad_array[i] = 0;
				}
				if (fwrite(pad_array, 1, pad_bytes, outfp) != pad_bytes) {
					fprintf(stderr, "read_and_convert: fwrite fails on padding\n");
					return(1);
				}
			}
		}

		/* close the input and output files */
		fclose(outfp);
		fclose(infp);

	}
					


	return(0);
}



	/*************************************************************
	 * PROCEDURE: decode_pixel_value
	 *
	 * DESCRIPTION: Given two bytes containing a single pixel
	 *              value from a FITS 16-bit integer file,
	 *              and the BZERO value from FITS header,
	 *              figure out the pixel value.  
	 *
	 *              Algorithm is:
	 *
	 *                  low-byte:   interpret all 8 bits as unsigned int
	 *                                   call this "A"
	 * 
 	 *                  high-byte:  interpret low 7 bits as unsigned int
	 *                                   call this "B"
	 *
	 *                              interpret most significant bit as
	 *                                   -32768    if bit = on
	 *                                        0    if bit = off
	 *                                   call this "C"
	 *
	 *              The final pixel value will be
	 *
	 *                  value =  (A + B + C) + bzero
	 *
	 * RETURNS:     pixel value
	 */


static int
decode_pixel_value
	(
	unsigned char low_byte,         /* I: low byte of pixel data */
	unsigned char high_byte,        /* I: high byte of pixel data */
	double bzero                    /* I: BZERO value from FITS header */
	)
{
	int low_val, high_val;
	int radix_val, pixel_val;

	low_val = (unsigned int) low_byte;
	high_val = (unsigned int) (high_byte & 0x7F);
	if (high_byte & 0x80) {
		radix_val = -32768;
	}
	else {
		radix_val = 0;
	}

	pixel_val = radix_val + high_val*256 + low_val;
	pixel_val += bzero;
#ifdef DEBUG2
	printf(" decode: low %5d  high %5d -> %5d   radix %6d   tot %6d \n",
				low_val, high_val, high_val*256, radix_val, pixel_val);
#endif

	return(pixel_val);
}


	/*************************************************************
	 * PROCEDURE: encode_pixel_value
	 *
	 * DESCRIPTION: Given a pixel value -- which must lie in the
	 *              range from 0 to +32767, since we have 
	 *              divided the original pixel value by 2 --
	 *              encode it into two bytes which can be written
	 *              into a 16-bit integer FITS file.
	 *
	 *              Algorithm is simple, since we know that the 
	 *              pixel value lies in range 0 to +32767.
	 *
	 *                  low_byte   = (value & 0x00FF)
	 * 
 	 *                  high-byte  = (value & 0x7F) 
	 *
	 *
	 *              Place the two bytes into the output arguments.
	 *
	 * RETURNS:     
	 *              0        if all goes well
	 *             -1        if an error occurs
	 */


static int
encode_pixel_value
	(
	int pixel_value,                /* I: encode this value */
	unsigned char *low_byte,        /* O: low byte of encoded pixel data */
	unsigned char *high_byte        /* O: high byte of encoded pixel data */
	)
{
	unsigned char low, high;

	/* sanity check */
	if ((pixel_value < 0) || (pixel_value > 32767)) {
		fprintf(stderr, "encode_pixel_value: invalid pixel value %d \n", 
					pixel_value);
		return(1);
	}

	low = (unsigned char) (pixel_value &  0x00FF);
	high = (unsigned char) ((pixel_value &  0x7F00) >> 8);
#ifdef DEBUG2
	printf(" encode: value  %6d  low %5d high %5d   \n",
				pixel_value, low, high);
#endif

	*low_byte = low;
	*high_byte = high;

	return(0);
}


	/*************************************************************
	 * PROCEDURE: pad_headerline     
	 *
	 * DESCRIPTION: Given a string which is intended for the FITS
	 *              header -- pad it with blank characters so that
	 *              it is 80 characters long.  
	 *
	 *              Do NOT place a NULL character at end!
	 * 
	 * RETURNS:     
	 *              0        if all goes well
	 *             -1        if an error occurs
	 */


static int
pad_headerline
	(
	char *headerline              /* I/O: characters to go in FITS header */
	)
{
	int i;

	/* 
	 * first, we walk forward until we reach the end of the 
	 *     'meaningful' portion of the line:
	 *
	 *            NULL '\0'     marks end
	 *            CARD_LENGTH   marks end
	 */
	for (i = 0; i < CARD_LENGTH; i++) {
		if (headerline[i] == '\0') {
			break;
		}
	}

	/* 
	 * now fill all entries from here to CARD_LENGTH with blank chars 
	 */
	for ( ; i < CARD_LENGTH; i++) {
		headerline[i] = ' ';
	}	

	return(0);
}


	/*************************************************************
	 * PROCEDURE: calc_data_block_size
	 *
	 * DESCRIPTION: Compute the size of each "data block" in the
	 *              FITS file.  That is, the size of the 
	 *              section of the file that contains the pixel data
	 *              for just one image.
	 *
	 *              It is NOT necessary to pad this "data block"
	 *              so that it has a multipled of 
	 *              (CARD_LENGTH * CARDS_PER_BUNCH) bytes.
	 *              The data for image (N+1) starts immediately
	 *              after the data for image N ends.
	 * 
	 * RETURNS:     
	 *              the size of the "data block" 
	 */


static int
calc_data_block_size
	(
	int naxis1,       /* I: number of columns of data in each image */
	int naxis2,       /* I: number of rows of data in each image */
	int bitpix        /* I: number of bits in each pixel (16 or -32) */
	)
{
	int i, bytes_per_pixel, pixdata_bytes;

#ifdef DEBUG
	printf("calc_data_block_size: naxis1 %d \n", naxis1);
	printf("calc_data_block_size: naxis2 %d \n", naxis2);
	printf("calc_data_block_size: bitpix %d \n", bitpix);
#endif

	if (bitpix == 16) {
		bytes_per_pixel = bitpix / 8;
	}
	else if (bitpix == -32) {
		bytes_per_pixel = (0 - bitpix) / 8;
	}
#ifdef DEBUG
	printf("calc_data_block_size: bytes_per_pixel = %d \n", bytes_per_pixel);
#endif

	/* the image contains this many bytes of pixel data */
	pixdata_bytes = naxis1 * naxis2 * bytes_per_pixel;
#ifdef DEBUG
	printf("   data_block_size = %d  \n", pixdata_bytes);
#endif

	return(pixdata_bytes);
}

