/**************************************************************************
 *                                                                        *
 * Copyright (c) 1996 Michael Richmond and Richard Treffers               *
 *                                                                        *
 *                    This software may be copied and distributed for     *
 *                    educational, research and not for profit services   *
 *                    provided that this copyright and statement are      *
 *                    included in all such copies.                        *
 *                                                                        *
 **************************************************************************/


/*	
	subtracts 'fileb' from 'filea', leaves result in 'filea'	

		SUB filea [fileb]  [const=c] [box=] [outfile=] [help]

	sub constant if const is specified

	1/29/91 - added 'floor' to expression for "const", <math.h>   - MWR
	9/9/92  - made arithmetic 32-bit, and added check against over/underflow
	          before converting back to 16-bit for output.  Now probably
	          quite a bit slower, unfortunately.   MWR
	9/22/92 - usage -rrt
	11/17/92 - outfile added -rrt
	1/10/96  - added 'help' option.  MWR
	7/25/2011 - added 'pmask' option.  MWR

	6/14/2013 - modified so if user specifies "outfile=", we copy
	                entire contents to new output file first,
	                then perform the requested subtraction.  MWR
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include "pcvista.h"
#include "fits.h"

#ifdef PROTO
int main(int, char **);
#endif

char *usage = "usage: sub file1 [file2] [const=] [box=] [outfile=] [pmask=] [help]";

int
main(argc, argv)
int argc;
char *argv[];
{
	FITS_HANDLE fh, fother = 0,fout, pmask_fh;
	static int i, constant, boxno;
	static int nr, nc, sr, sc, nrow, ncol, row, other_nrow, other_ncol;
	static int pmask_nrow, pmask_ncol;
	static int pmask_flag = 0;
	static int16 data[NMAX], other[NMAX], pmask_data[NMAX];
	long lresult[NMAX];
	char *gotit, other_file[NBUF],outfile[NBUF], pmask_file[NBUF];
	char err_msg[NBUF];
	int new_file=0;

	if (argc == 1) {
		error(-1, usage);
	}
	if (strcmp(argv[1], "help") == 0) {
		printf("%s\n", usage);
		exit(0);
	}

	fh = fits_open(argv[1], "r", &nrow, &ncol);
	strcpy(outfile,argv[1]);
	nr = nrow;
	nc = ncol;	

	for (i = 2; i < argc; i++) {
		if (keyword("help", argv[i])) {
			printf("%s\n", usage);
			exit(0);
		}
		if ((gotit = find("box", argv[i])) != NULL) {
			boxno = (int) evaluate(gotit);
			if (getbox(boxno, &sr, &sc, &nr, &nc))
				error(-1, "can't find box");
			check_box(boxno, nrow, ncol);			
			continue;
		}
		if ((gotit = find("outfile", argv[i])) != NULL) {
			strcpy(outfile,gotit);
			new_file=1;
			continue;
		}
		if ((gotit = find("pmask", argv[i])) != NULL) {
			strcpy(pmask_file,gotit);
			pmask_flag = 1;
			continue;
		}
		if ((gotit = find("const", argv[i])) != NULL) {
			constant = (int) floor(evaluate(gotit) + 0.5);
			continue;
		}
		fother = fits_open(argv[i], "r", &other_nrow, &other_ncol);
		strcpy(other_file, argv[i]);
		if (nrow != other_nrow)
			error(-1, "number of rows not the same");
		if (ncol != other_ncol)
			error(-1, "number of columns not the same");
	}

	if (pmask_flag == 1) {
		pmask_fh = fits_open(pmask_file, "r", &pmask_nrow, &pmask_ncol);
		if (nrow != pmask_nrow) {
			sprintf(err_msg, "number of rows in mask %s != number in %s \n",
						pmask_file, argv[1]);
			error(-1, err_msg);
		}
		if (ncol != pmask_ncol) {
			sprintf(err_msg, "number of cols in mask %s != number in %s \n",
						pmask_file, argv[1]);
			error(-1, err_msg);
		}
	}

	fout = fits_open(outfile, new_file?"w":"x", &nrow, &ncol);

	/*
	 * If the user specifies "outfile=", then we do two steps:
	 *   1.  copy original file into new output file, as-is
	 *   2.  then, operate on the new output file as directed by user.
	 * If we don't perform a complete copy first, and user requests
	 * action only within a box, then the regions outside the box
	 * won't have valid data copied into them.
	 */
	if (new_file) {
		fits_copyheader(fh, fout, FITS_CHECK);
		for (row = 0; row < nrow; row++) {
			fits_get_data(fh, row, 0, data, ncol);
			fits_put_data(fout, row, 0, data, ncol);
		}
	}
			
	for (row = sr; row < (sr + nr); row++) {
		fits_get_data(fh, row, sc, data, nc);
		if (pmask_flag == 1) {
			fits_get_data(pmask_fh, row, sc, pmask_data, nc);
		}
		for (i = 0; i < nc; i++) {
			lresult[i] = data[i];
		}
		if (fother) {
			fits_get_data(fother, row, sc, other, nc);
			for (i = 0; i < nc; i++) {
				if ((pmask_flag == 0) || (pmask_data[i] == 0))  {
					lresult[i] -= other[i];
				}
			}
		}
		if (constant) {
			for (i = 0; i < nc; i++) {
				if ((pmask_flag == 0) || (pmask_data[i] == 0)) {
					lresult[i] -= constant;
				}
			}
		}
		for (i = 0; i < nc; i++) {
			if (lresult[i] > MAX_DATA_VAL)
					data[i] = MAX_DATA_VAL;
				else if (lresult[i] < MIN_DATA_VAL)
					data[i] = MIN_DATA_VAL;
				else
					data[i] = (int16) lresult[i];
		}
		fits_put_data(fout, row, sc, data, nc);
	}
	fits_close(fout);
	if (pmask_flag == 1) {
		fits_close(pmask_fh);
	}
	exit(0);
}
