/**************************************************************************
 *                                                                        *
 * Copyright (c) 2016 Michael Richmond and Richard Treffers               *
 *                                                                        *
 *                    This software may be copied and distributed for     *
 *                    educational, research and not for profit services   *
 *                    provided that this copyright and statement are      *
 *                    included in all such copies.                        *
 *                                                                        *
 **************************************************************************/


/**********************************************************************
 * Compute a "bias correction region" for a single raw Tomoe image,
 *    and then subtract it from the image to create a "bias-subtracted" 
 *    version.
 *
 * The idea is taken from Ohsawa et al. 2016 
 *     SPIE 9913, Software and Cyberinfrastructure for Astronomy IV, 
 *             991339 (8 August 2016); doi: 10.1117/12.2231615
 *
 * 
 * MWR 10/19/2016
 **********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include "pcvista.h"
#include "fits.h"

#include <inttypes.h>
#include <time.h>
#include <sys/time.h>

#undef DEBUG



	/* 
	 * this structure holds the 'ngrid' consecutive rows of data
	 *    which were readout simultaneously.
	 */
typedef struct s_gslice {
	int id;            /* simple identifying value, for diagnostics */
	int num_row;       /* number of rows collected together here */
	int num_col;       /* number of columns in each row */
	int16 **data;      /* allocate for 'num_row' rows of 'num_col' pixels */
	int16 **pmask_data;   /* allocate for corresponding mask pixels */
} GSLICE;

typedef int (*PFI)();



#ifdef PROTO
int main(int, char **);

static GSLICE *
alloc_gslice(int bias_numrows, int ncol);

static int
free_gslice(GSLICE *slice);

static int 
fill_slices(FITS_HANDLE fh, FITS_HANDLE pmask_fh, int nrow, int ncol,
            int bias_startrow, int bias_numrows, int ngrid, 
            GSLICE **slices);

static GSLICE *
create_master_slice(int num_slice, GSLICE **slices);

static int 
write_master_slice(char *master_file, FITS_HANDLE master_fh, 
                   GSLICE *master_slice);

static int 
subtract_master_slice(FITS_HANDLE fh, FITS_HANDLE pmask_fh, FITS_HANDLE fout,
                      int nrow, int ncol, GSLICE *master_slice);

static int16
find_median(int num, int16 *values);

static int16
myselect(int num, int k, int16 *values);

static int
partition_array(int16 *values, int left_index, int right_index,
                int pivot_index);

#ifdef DEBUG
static void test_partition(void);
static void test_select(void);
static void print_array(int num, int16 *values);
static int compar_int16(int16 *a, int16 *b);
static int16 median_by_sort(int num, int16 *values);
static void test_median_speed(void);
#endif

#else
int main();
static GSLICE * alloc_gslice();
static int free_gslice();
static int fill_slices();
static GSLICE *create_master_slice();
static int write_master_slice();
static int subtract_master_slice();
static int16 find_median();
static int16 myselect();
static int partition_array();
#ifdef DEBUG
static void test_partition();
static void test_select();
static void print_array();
static int compar_int16();
static int16 median_by_sort();
static void test_median_speed();
#endif

#endif




static char err_msg[NBUF];

static char *usage = "usage: t_subbias file1 [sr=] [nr=] [ngrid=] [outfile=] [pmask=] [master=]  [help]";

int
main(argc, argv)
int argc;
char *argv[];
{
	FITS_HANDLE fh, fother = 0,fout, pmask_fh, master_fh;
	static int i, constant, boxno;
	static int nr, nc, sr, sc, nrow, ncol, row;
	static int pmask_nrow, pmask_ncol;
	static int master_nrow, master_ncol;
	static int pmask_flag = 0, master_flag = 0;
	static int bias_startrow, bias_numrows, ngrid, num_slice;
	static int16 data[NMAX], pmask_data[NMAX];
	long lresult[NMAX];
	char *gotit, outfile[NBUF], pmask_file[NBUF], master_file[NBUF];
	int new_file=0;
	GSLICE **gslice_array, *master_slice;

	if (argc == 1) {
		error(-1, usage);
	}
	if (strcmp(argv[1], "help") == 0) {
		printf("%s\n", usage);
		exit(0);
	}

#ifdef DEBUG
	{
		/* we usesome random numbers in DEBUG mode, so set seed here */
		time_t seed = time((time_t *) NULL);
		srandom(seed);
	}
#endif

	fh = fits_open(argv[1], "r", &nrow, &ncol);
	/* 
	 * initialize these FITS_HANDLES to an invalid value to indicate 
	 *    that the defaults are for 'no mask' and 'no output master slice'
	 */
	pmask_fh = -1;
	master_fh = -1;
	pmask_file[0] = '\0';
	master_file[0] = '\0';
	strcpy(outfile,argv[1]);
	nr = nrow;
	nc = ncol;	

	/* default values, which work on April 2016 Tomoe data */
	bias_numrows = 64;
	bias_startrow = 0;
	ngrid = 4;

	for (i = 2; i < argc; i++) {
		if (keyword("help", argv[i])) {
			printf("%s\n", usage);
			exit(0);
		}
		if ((gotit = find("nr", argv[i])) != NULL) {
			bias_numrows = (int) evaluate(gotit);
			if ((bias_numrows <= 0) || (bias_numrows >= nrow)) {
				sprintf(err_msg, "invalid number of bias rows %s \n", argv[i]);
				error(-1, err_msg);
			}
			continue;
		}
		if ((gotit = find("sr", argv[i])) != NULL) {
			bias_startrow = (int) evaluate(gotit);
			if ((bias_startrow < 0) || (bias_startrow >= nrow)) {
				sprintf(err_msg, "invalid startrow for bias %s \n", argv[i]);
				error(-1, err_msg);
			}
			continue;
		}
		if ((gotit = find("ngrid", argv[i])) != NULL) {
			ngrid = (int) evaluate(gotit);
			if ((ngrid <= 0) || (ngrid > bias_numrows)) {
				sprintf(err_msg, "invalid value for ngrid %s \n", argv[i]);
				error(-1, err_msg);
			}
			continue;
		}
		if ((gotit = find("outfile", argv[i])) != NULL) {
			strcpy(outfile,gotit);
			new_file=1;
			continue;
		}
		if ((gotit = find("pmask", argv[i])) != NULL) {
			strcpy(pmask_file,gotit);
			pmask_flag = 1;
			continue;
		}
		if ((gotit = find("master", argv[i])) != NULL) {
			strcpy(master_file,gotit);
			master_flag = 1;
			continue;
		}
	}
#ifdef DEBUG
	printf("t_subbias: bias_numrows    %d \n", bias_numrows);
	printf("t_subbias: bias_startrow   %d \n", bias_startrow);
	printf("t_subbias: ngrid           %d \n", ngrid);
	printf("t_subbias: outfile         %s \n", outfile);
	printf("t_subbias: master          %s \n", master_file);
#endif

	/* if the user specified a pmask file, open it up for reading */
	if (pmask_flag == 1) {
		pmask_fh = fits_open(pmask_file, "r", &pmask_nrow, &pmask_ncol);
		if (nrow != pmask_nrow) {
			sprintf(err_msg, "number of rows in mask %s != number in %s \n",
						pmask_file, argv[1]);
			error(-1, err_msg);
		}
		if (ncol != pmask_ncol) {
			sprintf(err_msg, "number of cols in mask %s != number in %s \n",
						pmask_file, argv[1]);
			error(-1, err_msg);
		}
	}

	/* if the user specified a 'master slice' file, open it for writing */
	if (master_flag == 1) {
		master_nrow = ngrid;
		master_ncol = ncol;
		master_fh = fits_open(master_file, "w", &master_nrow, &master_ncol);
	}

	/* 
	 * we may over-write the input FITS file with new data,
	 *     or create a new FITS file for the bias-subtracted output 
	 */
	fout = fits_open(outfile, new_file?"w":"x", &nrow, &ncol);




	/*
	 * If the user specifies "outfile=", then we do two steps:
	 *   1.  copy original file into new output file, as-is
	 *   2.  then, operate on the new output file as directed by user.
	 * If we don't perform a complete copy first, and user requests
	 * action only within a box, then the regions outside the box
	 * won't have valid data copied into them.
	 */
	if (new_file) {
		fits_copyheader(fh, fout, FITS_CHECK);
		for (row = 0; row < nrow; row++) {
			fits_get_data(fh, row, 0, data, ncol);
			fits_put_data(fout, row, 0, data, ncol);
		}
	}

	/* allocate for the needed GSLICE structures */
	num_slice = bias_numrows / ngrid;
	if ((gslice_array = (GSLICE **) malloc(num_slice*sizeof(GSLICE *))) 
			== NULL) {
		sprintf(err_msg, "malloc fails on gslice_array");
		error(1, err_msg);
	}
	for (i = 0; i < num_slice; i++) {
		if ((gslice_array[i] = alloc_gslice(ngrid, ncol)) == NULL) {
			sprintf(err_msg, "alloc_gslice fails for slice %d \n", i);
			error(1, err_msg);
		}
	}
#ifdef DEBUG
	printf("t_subbias: num_slice       %d \n", num_slice);
#endif


	/*
	 * Step 1: fill the slices with data from the bias region
	 */
	if (fill_slices(fh, pmask_fh, nrow, ncol, bias_startrow, bias_numrows, 
					ngrid, gslice_array) != 0) {
		sprintf(err_msg, "fill_slices fails");
		error(1, err_msg);
	}

#if 0
#ifdef DEBUG
	/* 
	 * a set of test routines, just for diagnostic purposes.
	 *    requires that DEBUG be set, too
	 */
	test_partition();
	test_select();
	test_median_speed();
	exit(0);
#endif
#endif

	/*
	 * Step 2: generate the "master slice"
	 */
	if ((master_slice = create_master_slice(num_slice, gslice_array)) == NULL) {
		sprintf(err_msg, "create_master_slice fails");
		error(1, err_msg);
	}
	/*
	 * If desired, write the master slice to a FITS file for later reference 
	 */
	if (master_flag == 1) {
		if (write_master_slice(master_file, master_fh, master_slice) != 0) {
			error(1, "write_master_slice fails");
		}
	}

	
	/*
	 * Step 3: subtract the "master slice" values from the input image,
	 *            using it repeatedly to tile the entire image.
	 */
	if (subtract_master_slice(fh, pmask_fh, fout, nrow, ncol, master_slice) != 0) {
		sprintf(err_msg, "create_master_slice fails");
		error(1, err_msg);
	}



	/*
	 * Step 4: close files, clean up
	 */
	fits_close(fh);
	fits_close(fout);
	if (pmask_flag == 1) {
		fits_close(pmask_fh);
	}

	/* de-allocate the GSLICE structures */
	for (i = 0; i < num_slice; i++) {
		if (free_gslice(gslice_array[i]) != 0) {
			sprintf(err_msg, "free_gslice fails");
			error(1, err_msg);
		}
	}
	if (free_gslice(master_slice) != 0) {
		sprintf(err_msg, "free_gslice fails on master_slice ");
		error(1, err_msg);
	}




	exit(0);
}


	/**************************************************************
	 * PROCEDURE: alloc_gslice
	 *
	 * DESCRIPTION: Allocate space for a single GSLICE structure.
	 *
	 * RETURNS:     A pointer to the GSLICE structure,  if all goes well
	 *              NULL                                if an error occurs
	 *
	 */

static GSLICE *
alloc_gslice
	(
	int num_rows,          /* I: number of rows in the 'slice' */
	int num_cols           /* I: number of columns in each row */
	)
{
	static int running_id = 0;
	int i;
	GSLICE *new_slice;

	if ((new_slice = (GSLICE *) malloc(sizeof(GSLICE))) == NULL) {
		sprintf(err_msg, "alloc_gslice: malloc fails for slice ");
		error(0, err_msg);
		return(NULL);
	}
	new_slice->id = running_id++;
	new_slice->num_row = num_rows;
	new_slice->num_col = num_cols;

	if ((new_slice->data = (int16 **) malloc(num_rows*sizeof(int16 *))) 
				== NULL) {
		sprintf(err_msg, "alloc_gslice: malloc fails for data pointers");
		error(0, err_msg);
		return(NULL);
	}
	if ((new_slice->pmask_data = (int16 **) malloc(num_rows*sizeof(int16 *))) 
				== NULL) {
		sprintf(err_msg, "alloc_gslice: malloc fails for pmask pointers");
		error(0, err_msg);
		return(NULL);
	}


	for (i = 0; i < num_rows; i++) {
		if ((new_slice->data[i] = (int16 *) calloc(num_cols, sizeof(int16))) 
				== NULL) {
			sprintf(err_msg, "alloc_gslice: calloc fails for data row %d", i);
			error(0, err_msg);
			return(NULL);
		}
		if ((new_slice->pmask_data[i] = (int16 *) calloc(num_cols, sizeof(int16))) 
				== NULL) {
			sprintf(err_msg, "alloc_gslice: calloc fails for pmask_data row %d", i);
			error(0, err_msg);
			return(NULL);
		}
	}

	return(new_slice);
}


	/**************************************************************
	 * PROCEDURE: free_gslice
	 *
	 * DESCRIPTION: De-allocate all the memory used by a GSLICE
	 *              structure, and then the structure itelf.
	 *
	 * RETURNS:     0          if all goes well
	 *              1          if an error occurs
	 *
	 */

int 
free_gslice
	(
	GSLICE *slice        /* I/O: structure to destroy */
	)
{
	int i;

	for (i = 0; i < slice->num_row; i++) {
		free(slice->data[i]);
		free(slice->pmask_data[i]);
	}
	free(slice);
	
	return(0);
}



	/**************************************************************
	 * PROCEDURE: fill_slices
	 *
	 * DESCRIPTION: Read data from the bias region of the input file
	 *              and store it in the appropriate elements
	 *              of the GSLICE array given as final argument.
	 *
	 * RETURNS:     0        if all goes well
	 *              1        if an error occurs
	 *
	 */

static int 
fill_slices
	(
	FITS_HANDLE fh,        /* I: handle to input FITS image */
	FITS_HANDLE pmask_fh,  /* I: handle to FITS file for mask */
	int nrow,              /* I: number of total rows in input image */
	int ncol,              /* I: number of total cols in input image */
	int bias_startrow,     /* I: starting row of the bias region */
	int bias_numrows,      /* I: number of rows in the bias region */
	int ngrid,             /* I: number of consecutive rows in a slice */
	GSLICE **slices        /* O: array of GSLICE into which we put data */
	)
{
	int row, num_slice, slice_index, slice_r;
	int start_col = 0;
	GSLICE *this_slice;

	num_slice = bias_numrows / ngrid;
	row = bias_startrow;

	for (slice_index = 0; slice_index < num_slice; slice_index++) {

		this_slice = slices[slice_index];

		for (slice_r = 0; slice_r < this_slice->num_row; slice_r++) {

			fits_get_data(fh, row, start_col, 
						this_slice->data[slice_r], ncol);
#ifdef DEBUG
			printf(" slice %5d r %5d  = row %5d :  %6d %6d %6d \n",
					slice_index, slice_r, row, 
					this_slice->data[slice_r][0],
					this_slice->data[slice_r][1],
					this_slice->data[slice_r][2]);
#endif

			if (pmask_fh != -1) {
				fits_get_data(pmask_fh, row, start_col, 
						this_slice->pmask_data[slice_r], ncol);
#ifdef DEBUG
				printf(" mask slice %5d r %5d  = row %5d :  %6d %6d %6d \n",
					slice_index, slice_r, row, 
					this_slice->pmask_data[slice_r][0],
					this_slice->pmask_data[slice_r][1],
					this_slice->pmask_data[slice_r][2]);
#endif
			}

			row++;
		}
	}

	return(0);
}




	/**************************************************************
	 * PROCEDURE: create_master_slice
	 *
	 * DESCRIPTION: Given a set of GSLICE structures, filled
	 *              with data from the bias region of the input 
	 *              image, create a 'master' slice which will
	 *              serve as the template for our eventual 
	 *              bias subtraction.
	 *
	 *              We go pixel-by-pixel, gathering together the
	 *              'num_slices' pixels in identical locations
	 *              in all the slices.  We discard any values
	 *              for which the pmask_data value is non-zero.
	 *              Then, we take the median of all surviving
	 *              pixel values, and use that as the 'master' value.
	 *              
	 *
	 * RETURNS:     pointer to master slice     if all goes well
	 *              NULL                        if an error occurs
	 *
	 */

static GSLICE *
create_master_slice
	(
	int num_slice,         /* I: number of slices to analyze */
	GSLICE **slices        /* I: array of GSLICE structs with input data */
	)
{
	int nrow, ncol, row, col;
	int slice_index;
	int16 *pixval_array;
	GSLICE *master_slice, *this_slice;

	/* 
	 * read number of rows and cols of pixels from the first slice,
	 *    just for convenience.  If we were paranoid, we'd check
	 *    to see if all slices had identical 'nrow' and 'ncol' values.
	 */
	nrow = slices[0]->num_row;
	ncol = slices[0]->num_col;

	/* 
	 * allocate for an array of int16 values -- which will contain
	 *    on value per slice.  We'll use this array to compute median
	 */
	if ((pixval_array = (int16 *) malloc(num_slice*sizeof(int16))) == NULL) {
		sprintf(err_msg, "create_master_slice: can't malloc for pixval_array");
		error(0, err_msg);
		return((GSLICE *) NULL);
	}

	/* allocate for the return GSLICE structure */
	if ((master_slice = (GSLICE *) alloc_gslice(nrow, ncol)) == NULL) {
		sprintf(err_msg, "create_master_slice: can't malloc for master_slice");
		error(0, err_msg);
		return((GSLICE *) NULL);
	}


	/* 
	 * now, begin the main loop, finding median of pixel values 
	 *     at each location
	 */
	for (row = 0; row < nrow; row++) {
		for (col = 0; col < ncol; col++) {

			/* 
			 * this is the number of pixel values from which to calc
			 *    the median.  It will usually be = 'num_slice', 
			 *    but could be smaller if some pixels are masked out.
		 	 */
			int num_pix = 0;
			int median_val;

			for (slice_index = 0; slice_index < num_slice; slice_index++) {
				int16 this_pixval;

				this_slice = slices[slice_index];
				if (this_slice->pmask_data[row][col] != 0) {
#ifdef DEBUG
					printf(" row %5d col %5d slice %5d =  MASK \n",
							row, col, slice_index);
#endif
					continue;
				}

				this_pixval = this_slice->data[row][col];
				pixval_array[slice_index] = this_pixval;
#ifdef DEBUG
				printf(" row %5d col %5d slice %5d =  %6d \n",
					row, col, slice_index, this_pixval);
#endif
				num_pix++;
			}

			/* 
			 * okay, we have all the relevant pixel values in the 
			 *    pixval_array[].  Now we compute median
			 */
			median_val = find_median(num_pix, pixval_array);

			/* set this pixel's value in the master slice */
			master_slice->data[row][col] = median_val;			
		}
	}

	/* free the space for the structures we've allocated */
	free(pixval_array);

	return(master_slice);
}



	/**************************************************************
	 * PROCEDURE: write_master_slice
	 *
	 * DESCRIPTION: Write the data in the 'master_slice' to a
	 *              FITS file.   Write the 'pmask' also.
	 *              
	 *
	 * RETURNS:     0        if all goes well
	 *              1        if an error occurs
	 *
	 */

static int 
write_master_slice
	(
	char *master_file,       /* I: name of output FITS file for the slice */
	FITS_HANDLE master_fh,   /* I: handle to input FITS image */
	GSLICE *master_slice     /* I: array of GSLICE with the data */
	)
{
	int row, num_slice, slice_index, slice_r;
	int mask_nrow, mask_ncol;
	int start_col = 0;
	char pmask_filename[NBUF];
	FITS_HANDLE master_pmask_fh;
	GSLICE *this_slice;

	/* we've already opened the master_fh, so we can just write the data */
	for (row = 0; row < master_slice->num_row; row++) {
		fits_put_data(master_fh, row, start_col, master_slice->data[row], 
						master_slice->num_col);
	}
	fits_close(master_fh);

	/* 
	 * create a pmask for the master slice file.  We modify the name
	 *   of the master_file so that it ends with .msk
	 */
	mask_nrow = master_slice->num_row;
	mask_ncol = master_slice->num_col;
	sprintf(pmask_filename, "%s.msk", master_file);
	master_pmask_fh = fits_open(pmask_filename, "w", &mask_nrow, &mask_ncol);
	for (row = 0; row < master_slice->num_row; row++) {
		fits_put_data(master_pmask_fh, row, start_col, 
						master_slice->pmask_data[row], master_slice->num_col);
	}
	fits_close(master_pmask_fh);

	return(0);
}




	/**************************************************************
	 * PROCEDURE: subtract_master_slice
	 *
	 * DESCRIPTION: Use the given "master" GSLICE structure
	 *              as a template for the bias.  Tile it over the
	 *              given image and subtract it from the pixel values.
	 *
	 *              If the 'pmask' value is set for a pixel, do nothing.
	 *
	 * RETURNS:     0        if all goes well
	 *              1        if an error occurs
	 *
	 */

static int 
subtract_master_slice
	(
	FITS_HANDLE fh,        /* I: handle to input FITS image */
	FITS_HANDLE pmask_fh,  /* I: handle to input FITS file for mask */
	FITS_HANDLE fout,      /* I: handle to output FITS image */
	int nrow,              /* I: number of total rows in input image */
	int ncol,              /* I: number of total cols in input image */
	GSLICE *master_slice   /* I: GSLICE holding bias values */
	)
{
	int row, col, num_slice, slice_index, slice_r;
	int num_tile, tile_index;
	int start_col = 0;
	int16 image_data[NMAX], pmask_data[NMAX];
	long lresult;
	GSLICE *this_slice;

	/* 
	 * Compute how many times must we replicate the master slice 
	 *    to cover the entire image.  Note that we will leave
	 *    any "leftover" rows untouched.
	 */
	num_tile = nrow / (master_slice->num_row);

	row = 0;
	start_col = 0;
	for (tile_index = 0; tile_index < num_tile; tile_index++) {
#ifdef DEBUG
		printf("subtract_master_slice: start tile_index %d \n", tile_index);
#endif

		for (slice_r = 0; slice_r < master_slice->num_row; slice_r++) {

			fits_get_data(fh, row, start_col, image_data, ncol);
#ifdef DEBUG
			printf("      r %5d  = row %5d :  %6d %6d %6d \n",
					slice_r, row, 
					image_data[0], image_data[1], image_data[2]);
#endif

			if (pmask_fh != -1) {
				fits_get_data(pmask_fh, row, start_col, pmask_data, ncol);
#ifdef DEBUG
				printf(" mask r %5d  = row %5d :  %6d %6d %6d \n",
					slice_r, row, 
					pmask_data[0], pmask_data[1], pmask_data[2]);
#endif
			}

			for (col = start_col; col < ncol; col++) {

				/* 
				 * check both the master_slice's pmask value and the
				 *     image's pmask_value.  If either one is non-zero,
				 *     then we do NOT subtract the bias value from this pixel
				 */
				if ((pmask_data[col] != 0) || 
						(master_slice->pmask_data[slice_r][col] != 0)) {
					continue;
				}

				/* otherwise, subtract the bias value */
				lresult = image_data[col];
				lresult -= master_slice->data[slice_r][col];

				/* make sure that the result fits into the permitted int16 range */
				if (lresult > MAX_DATA_VAL) {
					image_data[col] = MAX_DATA_VAL;
				}
				else if (lresult < MIN_DATA_VAL) {
					image_data[col] = MIN_DATA_VAL;
				}
				else {
					image_data[col] = lresult;
				}
			}

			/* 
			 * Okay, we can write this entire row of bias-subtracted data 
			 *      to the output file now */
			fits_put_data(fout, row, start_col, image_data, ncol);

			row++;
		}
	}

	return(0);
}





	/**************************************************************
	 * PROCEDURE: find_median
	 *
	 * DESCRIPTION: Given an array of 'num' int16 values,
	 *              find the median.
	 *
	 *              If 'num' is even, we return the lower of the
	 *              two middle values.  If 'num' is odd, we return
	 *              the middle value.
	 *
	 *              This implementation uses the 'quick_select'
	 *              algorithm.
	 *              
	 * RETURNS:     median value 
	 *
	 */

static int16
find_median
	(
	int num,               /* I: number of values to search */
	int16 *values          /* I: array of 'num' pixel values */
	)
{
	int median_index;
	int16 result;

	/* we want the item with the median_index'th largest value */
	median_index = num / 2;
	
	result = myselect(num, median_index, values);

#ifdef DEBUG2
	printf("find_median: num %5d  median_index %5d   result %5d \n",
				num, median_index, result);
#endif

	return(result);
}


	/**************************************************************
	 * PROCEDURE: myselect
	 *
	 * DESCRIPTION: Given a pointer to an array of 'num' int16
	 *              values, and an index 'k' indicating
	 *              that we want the k'th largest value in the array,
	 *              use a recursive splitting technique to find
	 *              that value.
	 *              
	 * RETURNS:     the k'th largest value in the array
	 *
	 */

static int16
myselect
	(
	int num,               /* I: number of values in the array */
	int k,                 /* I: we want the k'th largest value */
	int16 *values          /* I: array of 'num' pixel values */
	)
{
	int left_index, right_index, pivot_index;
	
	left_index = 0;
	right_index = num - 1;

	do {
		if (left_index == right_index) {
			return(values[left_index]);
		}

		/* now pick some index between left and right */
		pivot_index = (left_index + right_index) / 2;

		/* 
		 * re-arrange the array values into sections with values
		 *   less than that of the pivot_index and greater than
		 *   that of the pivot_index
	 	 */
		pivot_index = partition_array(values, left_index, right_index,
					pivot_index);
	
		if (k == pivot_index) {
			return(values[pivot_index]);
		}
		else if (k < pivot_index) {
			right_index = pivot_index - 1;
		}
		else {
			left_index = pivot_index + 1;
		}
	} while (left_index < right_index);

	/* if we get here, then the median must be the left_index'th value */
	return(values[left_index]);
}





	/**************************************************************
	 * PROCEDURE: partition_array
	 *
	 * DESCRIPTION: Given a pointer to an array of 'num' int16
	 *              values, and indices marking
	 *
	 *                 the left edge of the interesting section
	 *                 the right edge of the interesting section
	 *                 some 'pivot' element within the interesting region
	 *
	 *              this function will re-arrange elements within
	 *              the array so that we end up with
	 *        
	 *                 left  portion has values <  orig pivot value
	 *                 right portion has values >= orig pivot value
	 *
	 * RETURNS:     the final location of the pivot value
	 *
	 */

#define SWAP(a, b)   { int16 temp = (a); (a) = (b); (b) = temp; }

static int
partition_array
	(
	int16 *values,         /* I: array of 'num' pixel values */
	int left_index,        /* I: left edge of interesting region */
	int right_index,       /* I: right edge of interesting region */
	int pivot_index        /* I: index we use as comparison in re-arranging */
	)
{
	int i, store_index;
	int16 pivot_value;

	pivot_value = values[pivot_index];


#ifdef DEBUG2
		printf(" partition_array: left %5d right %5d pivot %5d = %5d \n",
				left_index, right_index, pivot_index, pivot_value);
		print_array((right_index - left_index) + 1, values + left_index);
#endif


	/* move the pivot value to right edge, for convenience */
	SWAP(values[right_index], values[pivot_index]);
	store_index = left_index;
#ifdef DEBUG2
		printf(" after initial swap %3d = %5d and %3d = %5d \n",
				right_index, values[pivot_index], pivot_index, values[right_index]);
		print_array((right_index - left_index) + 1, values + left_index);
#endif


	for (i = left_index; i < right_index; i++) {
		if (values[i] < pivot_value) {
			SWAP(values[store_index], values[i]);
			store_index++;
#ifdef DEBUG2
				printf(" after swapping %3d = %5d and %3d = %5d \n",
								store_index, values[i], i, values[store_index]);
				print_array((right_index - left_index) + 1, values + left_index);
#endif
		}
	}
	SWAP(values[right_index], values[store_index]);
#ifdef DEBUG2
	printf(" after final swap %3d = %5d and %3d = %5d \n",
			store_index, values[right_index], right_index, values[store_index]);
	print_array((right_index - left_index) + 1, values + left_index);
#endif


#ifdef DEBUG2
	printf(" after  paritioning \n");
	print_array((right_index - left_index) + 1, values + left_index);
#endif

	return(store_index);
}


#ifdef DEBUG

	/**************************************************************
	 * PROCEDURE: test_partition
	 *
	 * DESCRIPTION: A procedure to test the 'partition_array'
	 *              function, for debugging.
	 *
	 * RETURNS:     nothing
	 *
	 */


#define TEST_NUM  16
#define TEST_ITER  100000
#define TEST_ARRAYS 100000

static void
test_partition
	(
	void
	)
{
	int i, max, retval;
	long randval;
	int16 values[TEST_NUM];

	for (i = 0; i < TEST_NUM; i++) {
		/* make sure the random value lies in the range of int16 */
		randval = random() / 131072;
		values[i] = (int16) randval;
	}

	/* now call the parition function */
	
	retval = partition_array(values, 0, 8, 4);	

	return;
}



static void
test_select
	(
	void
	)
{
	int iter, median_index;
	int i, max, my_retval, sort_retval;
	long randval;
	int16 values[TEST_NUM];

	median_index = TEST_NUM / 2;

	for (iter = 0; iter < TEST_ITER; iter++) {

			  for (i = 0; i < TEST_NUM; i++) {
					/* make sure the random value lies in the range of int16 */
					randval = random() / 131072;
				  values[i] = (int16) randval;
			  }

#ifdef DEBUG2
			  printf(" test_select: ");
				print_array(TEST_NUM, values);
#endif

			  /* find median using 'myselect' method */
			  my_retval = myselect(TEST_NUM, median_index, values);
#ifdef DEBUG2
			  printf(" test_select: my_retval is %5d \n", my_retval);
#endif

			/* find median by sorting method */
			sort_retval = median_by_sort(TEST_NUM, values);
#ifdef DEBUG2
			printf(" test_select: sort_retval is %5d \n", sort_retval);
#endif

			/* are they the same? print error message if not */
			if (my_retval != sort_retval) {
				printf(" my_retval %5d NOT sort_retval %5d \n", my_retval, sort_retval);
			}

	}

	return;
}




	/**************************************************************
	 * PROCEDURE: print_array
	 *
	 * DESCRIPTION: Print out an array of integers in a nice format.
	 *              Useful for debugging.
	 *
	 * RETURNS:     nothing
	 *
	 */

static void
print_array
	(
	int num,               /* I: number of elements in array */
	int16 *values          /* I: array of 'num' values to print */
	)
{
	int i;

	for (i = 0; i < num; i++) {
		printf(" %6d", values[i]);
	}
	printf("\n");
}


	/**************************************************************
	 * PROCEDURE: compar_int16
	 *
	 * DESCRIPTION: Compare two int16 values, return a value
	 *              indicating which is larger.
	 * 
	 *              Used in sorting.
	 *
	 * RETURNS:     -1         if a < b
	 *               0         if a == b
	 *               1         if a > b
	 *
	 */

static int
compar_int16
	(
	int16 *a,              /* I: compare this int16 value ...  */
	int16 *b               /* I: ... to this one */
	)
{
	if ((*a) < (*b)) {
		return(-1);
	}
	if ((*a) > (*b)) {
		return(1);
	}
	return(0);
}


	/**************************************************************
	 * PROCEDURE: sort_array
	 *
	 * DESCRIPTION: Sort an array of int16 values using quicksort.
	 *
	 * RETURNS:     nothing
	 *
	 */

static void
sort_array
	(
	int num,               /* I: number of elements in array */
	int16 *values          /* I: array of 'num' values to sort */
	)
{
	qsort(values, num, sizeof(int16), (PFI) compar_int16);
}

	/**************************************************************
	 * PROCEDURE: median_by_sort
	 *
	 * DESCRIPTION: Find the median value in an array of pixel values
	 *              by calling qsort(), and then picking the middle element.
	 *
	 * RETURNS:     median value of array
	 *
	 */

static int16
median_by_sort
	(
	int num,               /* I: number of elements in array */
	int16 *values          /* I: array of 'num' values in which to find median */
	)
{
	int median_index;
	int16 median_value;

	median_index = num / 2;

	sort_array(num, values);
	median_value = values[median_index];
	
	return(median_value);
}




	/**************************************************************
	 * PROCEDURE: test_median_speed
	 *
	 * DESCRIPTION: Compare the speed of routines to find the median
	 *              value of an array of numbers.
	 *
	 * RETURNS:     nothing
	 *
	 */

static void
test_median_speed
	(
	void
	)
{
	int iter, median_index, array_index;
	int i, j, max, my_retval, sort_retval;
	long randval;
	int16 **values;

	long us;    /* time in microseconds */
	time_t s;   /* time in seconds */
	struct timeval tv;
	double start_t, end_t, t1, t2;

	
	/* 
	 * prepare TEST_ARRAYS arrays of random numbers; 
	 *     we'll use them repeatedly if necessary.  This number of arrays
	 *     should be large enough to provide a sufficient variety
	 *     of arrangements of numbers.
	 */
	if ((values = (int16 **) malloc(TEST_ARRAYS*sizeof(int16 *))) == NULL) {
		fprintf(stderr, "can't malloc for values pointers \n");
		exit(1);
	}
	for (i = 0; i < TEST_ARRAYS; i++) {
		if ((values[i] = (int16 *) malloc(TEST_NUM*sizeof(int16))) == NULL) {
			fprintf(stderr, "can't malloc for values [%5d] of %d int16 \n",
						i, TEST_NUM);
			exit(1);
		}
	}
	for (i = 0; i < TEST_ARRAYS; i++) {
		for (j = 0; j < TEST_NUM; j++) {
			/* make sure the random value lies in the range of int16 */
			randval = random() / 131072;
			values[i][j] = (int16) randval;
		}
	}


	median_index = TEST_NUM / 2;

	/* first, the 'select' method */

	gettimeofday(&tv, NULL);
	s = tv.tv_sec;
	us = tv.tv_usec;
	start_t = (double) s + ((double) us / 1.0e6);
	array_index = 0;
	for (iter = 0; iter < TEST_ITER; iter++) {
		my_retval = myselect(TEST_NUM, median_index, values[array_index]);
		if (++array_index >= TEST_ARRAYS) {
			array_index = 0;
		}
	}
	gettimeofday(&tv, NULL);
	s = tv.tv_sec;
	us = tv.tv_usec;
	end_t = (double) s + ((double) us / 1.0e6);
	
	t1 = end_t - start_t;
	printf(" select method %19.10f \n", t1);



	/* next, the 'sort' method */

	gettimeofday(&tv, NULL);
	s = tv.tv_sec;
	us = tv.tv_usec;
	start_t = (double) s + ((double) us / 1.0e6);
	array_index = 0;
	for (iter = 0; iter < TEST_ITER; iter++) {
		/* find median using 'myselect' method */
		my_retval = median_by_sort(TEST_NUM, values[array_index]);
		if (++array_index >= TEST_ARRAYS) {
			array_index = 0;
		}
	}
	gettimeofday(&tv, NULL);
	s = tv.tv_sec;
	us = tv.tv_usec;
	end_t = (double) s + ((double) us / 1.0e6);
	
	t2 = end_t - start_t;
	printf(" sort method %19.10f \n", t2);


	return;
}



#endif
