
/**************************************************************************
 *                                                                        *
 * Copyright (c) 1996 Michael Richmond and Richard Treffers               *
 *                                                                        *
 *                    This software may be copied and distributed for     *
 *                    educational, research and not for profit services   *
 *                    provided that this copyright and statement are      *
 *                    included in all such copies.                        *
 *                                                                        *
 **************************************************************************/


/*	
	Transposes an image, so an input MxN image turns into an output NxM image.
	All pixel values remain unchanged.

	By default, the original image is overwritten.  In order to preserve
	the original file, use the "outfile=" command-line option.

	12/30/2017 MWR
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include "pcvista.h"
#include "fits.h"

#ifdef PROTO
int main(int, char **);
static int transpose_image(void);
#else
static int transpose_image();
#endif

static int ncol;
static int nrow;
static FITS_HANDLE fin, fout;
static int16 data1[NMAX], data2[NMAX];

#define USAGE "usage: transpose image.fit  [outfile=] [help]"
	/* 
	 * this is appended to the input file name for use as
	 *    a unique and unused (we hope) file name.  If it is 
	 *    created, we'll delete it before terminating the program.
	 *
	 * So, for example, "vega.fits" becomes "vega.fits_XX_transpose_XX_"
	 */
#define TEMP_EXTENSION "_XX_transpose_XX_"

int
main(argc,argv)
int argc;
char *argv[];
{
	int i;
	char *gotit,outfile[NBUF];
	char err_msg[NBUF];
	int new_file = 0;

	if (argc < 2) {
		error(-1, USAGE);
	}
	if (strcmp(argv[1], "help") == 0) {
		printf("%s\n", USAGE);
		exit(0);
	}


	fin = fits_open(argv[1], "r", &nrow, &ncol);
	/* 
	 * we create a temporary file that has the same name as the 
	 *   input file, but with an added extension that we 
	 *   hope will not conflict with any existing file.
	 */
	sprintf(outfile, "%s%s", argv[1], TEMP_EXTENSION);

	for (i = 2; i < argc; i++){
		if (keyword("help",argv[i])) {
			printf("%s\n",USAGE);
			exit(0);
		}
		if ((gotit = find("outfile", argv[i])) != NULL) {
			strcpy(outfile,gotit);
			new_file = 1;
			continue;
		}
		fprintf(stderr,"transpose: unknown option '%s'\n",argv[i]);
		exit(1);
	}

	fout = fits_open(outfile, "w", &ncol, &nrow);
		 
	if(new_file) {
		fits_copyheader(fin,fout, FITS_NOCHECK);
	}

	/* this routine does all the work */
	if (transpose_image() != 0) {
		error(1, "transpose: transpose_image returns with error");
	}
	
	fits_close(fout);

	/* now, if necessary, replace the original with the transposed copy */
	if (new_file == 0) {
		if (unlink(argv[1]) != 0) {
			sprintf(err_msg, "transpose: failure to unlink orig %s", 
					argv[1]);
			error(1, err_msg);
		}
		if (rename(outfile, argv[1]) != 0) {
			int my_errno = errno;
			sprintf(err_msg, "transpose: failure to rename temp file %s", 
					outfile);
			fprintf(stderr, "errno is %d \n", my_errno);
			fprintf(stderr, "%s \n", strerror(my_errno));
			error(1, err_msg);
		}
	}

	exit(0);
}

static int
transpose_image
	(
	void
	)
{
	int row, col;
	int16 **row_pix;
	int16 new_pix[NMAX];

	/* first, allocate space for an entire copy of the original image */
	if ((row_pix = (int16 **) calloc(nrow, sizeof(int16 *))) == NULL) {
		error(1, "transpose: calloc fails to get memory for row_pix");
	}
	for (row = 0; row < nrow; row++) {
		if ((row_pix[row] = (int16 *)calloc(ncol, sizeof(int16))) == NULL) {
			error(1, "transpose: calloc fails to get memory for row_pix[]");
		}
	}

	/* copy all pixels from the original image into memory */
	for (row = 0; row < nrow; row++) {
		fits_get_data(fin, row, 0, row_pix[row], ncol);
	}
		
	/* 
	 * now, walk through this image in memory by COLS, instead of ROWS,
	 *    and build up one transposed row at a time; write that row
	 *    to the output file; then go to the next transposed row.
	 */
	for (col = 0; col < ncol; col++) {
		for (row = 0; row < nrow; row++) {
			new_pix[row] = row_pix[row][col];
		}
		fits_put_data(fout, col, 0, new_pix, nrow);
	}

	/* de-allocate the memory, since we've written the copy to disk */
	for (row = 0; row < nrow; row++) {
		free(row_pix[row]);
	}

	return(0);
}

